# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from inspect import signature

from qtpy.QtCore import QSize

from karabo.common.api import WeakMethodRef
from karabo.native import (
    AccessLevel, AccessMode, Configurable, Hash, Int32, Schema, Timestamp)
from karabogui import access as krb_access
from karabogui.binding.api import (
    BindingRoot, DeviceClassProxy, DeviceProxy, ProjectDeviceProxy,
    ProxyStatus, build_binding)
from karabogui.events import KaraboEvent
from karabogui.testing import singletons, system_hash
from karabogui.topology.system_topology import SystemTopology

from ..manager import Manager, project_db_handler
from ..mediator import Mediator

TEST_SERVER_ID = 'swerver'
TEST_CLASS_ID = 'PrettyDevice'
TEST_DEVICE_ID = 'dev'


def make_project_db_handler(fall_through):
    # Creates a wrapped function to test the project_db_handler decorator
    @project_db_handler(fall_through)
    def handle(self, success, request, reply, reason=''):
        return reply

    return handle


class PrettyDevice(Configurable):
    init_prop = Int32(accessMode=AccessMode.INITONLY)
    ro_prop = Int32(accessMode=AccessMode.READONLY)


def _get_class_proxy(schema):
    binding = build_binding(schema)
    return DeviceClassProxy(binding=binding)


def _get_project_device_proxy(server_id, class_id, device_id):
    with singletons(topology=SystemTopology()):
        binding = BindingRoot(class_id=class_id)
        proxy = ProjectDeviceProxy(device_id=device_id,
                                   server_id=server_id,
                                   binding=binding)
        build_binding(PrettyDevice.getClassSchema(), existing=proxy.binding)
    return proxy


def test_shutdown_device(mocker):
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        manager.shutdownDevice('dev', showConfirm=False)
        network.onKillDevice.assert_called_with('dev')
        network.reset_mock()
        manager.shutdownDevice('dev', showConfirm=False)
        mb = mocker.patch('karabogui.singletons.manager.QMessageBox')
        mb().size.return_value = QSize(10, 10)
        mb().exec.return_value = mb.Yes
        manager.shutdownDevice('dev')
        network.onKillDevice.assert_called_with('dev')
        network.onKillDevice.reset_mock()
        network.reset_mock()
        mb.reset_mock()
        manager.shutdownDevice('dev', showConfirm=False)
        network.onKillDevice.assert_called_with('dev')
        network.onKillDevice.reset_mock()
        mb().exec.return_value = mb.No
        manager.shutdownDevice('dev')
        network.onKillDevice.assert_not_called()

        mb = mocker.patch('karabogui.singletons.manager.QMessageBox')
        mb().size.return_value = QSize(10, 10)
        mb().exec.return_value = mb.Yes
        network.reset_mock()
        manager.shutdownDevice('dev')
        network.onKillDevice.assert_called_with('dev')

        mb().exec.return_value = mb.No
        network.reset_mock()
        manager.shutdownDevice('dev')
        network.onKillDevice.assert_not_called()


def test_shutdown_server(mocker):
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        mb = mocker.patch('karabogui.singletons.manager.QMessageBox')
        mb().size.return_value = QSize(10, 10)
        mb().exec.return_value = mb.Yes
        manager.shutdownServer('swerver')
        network.onKillServer.assert_called_with('swerver')
        network.onKillServer.reset_mock()
        mb.reset_mock()
        mb().exec.return_value = mb.No
        manager.shutdownServer('swerver')
        network.onKillServer.assert_not_called()

        mb().exec.return_value = mb.No
        manager.shutdownServer('swerver')
        network.onKillServer.assert_not_called()


def test_call_device_slot(mocker):
    handler_args = None

    def handler(success, reply):
        nonlocal handler_args
        handler_args = success, reply

    def request_handler(success, reply, request):
        nonlocal handler_args
        handler_args = success, reply, request

    params = Hash('arg0', 0, 'arg1', 1)
    reply = Hash('result', 'yer dumb')

    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()

        # 1. Call the slot of the 'remote device'
        token = manager.callDeviceSlot(handler, 'dev', 'slot', params)

        assert token in manager._request_handlers
        network.onExecuteGeneric.assert_called_with(
            'dev', 'slot', params, token=token)

        # Pretend the reply arrived
        manager.handle_requestGeneric(success=True,
                                      request=Hash("token", token),
                                      reply=reply)
        assert handler_args == (True, reply)

        # 2. Different signature, we receive request
        token = manager.callDeviceSlot(request_handler,
                                       'dev', 'slot', params)

        assert token in manager._request_handlers
        network.onExecuteGeneric.assert_called_with(
            'dev', 'slot', params, token=token)

        manager.handle_requestGeneric(success=True,
                                      request=Hash("token", token),
                                      reply=reply)
        assert handler_args == (True, reply, Hash("token", token))

        # 3. Test with WeakMethod

        class TestObject:
            def handler(self, success, reply, request):
                nonlocal handler_args
                handler_args = success, reply, request

        obj = TestObject()
        weak_handler = WeakMethodRef(obj.handler, num_args=3)
        sig = signature(weak_handler)
        assert len(sig.parameters) == 3

        token = manager.callDeviceSlot(weak_handler,
                                       'dev', 'slot', params)

        assert token in manager._request_handlers
        network.onExecuteGeneric.assert_called_with(
            'dev', 'slot', params, token=token)
        manager.handle_requestGeneric(success=True,
                                      request=Hash("token", token),
                                      reply=reply)
        assert handler_args == (True, reply, Hash("token", token))


def test_handle_system_topology(gui_app, mocker):
    topo_hash = system_hash()
    topology = mocker.Mock()
    mediator = Mediator()
    with singletons(topology=topology, mediator=mediator):
        manager = Manager()
        manager.handle_systemTopology(topo_hash)
        topology.initialize.assert_called_with(topo_hash)


def test_handle_instance_new(mocker):
    network = mocker.Mock()
    topology = SystemTopology()
    topo_hash = system_hash()
    topology.initialize(topo_hash)
    with singletons(network=network, topology=topology):
        manager = Manager()
        target = 'karabogui.singletons.manager.broadcast_event'
        broadcast_event = mocker.patch(target)
        topo_hash = system_hash()
        del topo_hash['device.divvy']
        del topo_hash['macro']

        changes = Hash("new", topo_hash,
                       "update", Hash(), "gone", Hash())
        manager.handle_topologyUpdate(changes)

        topo_update = {
            'devices': [('orphan', 'Parentless',
                         ProxyStatus.ONLINE)],
            'servers': [('swerver', 'BIG_IRON', ProxyStatus.ONLINE)]
        }
        calls = broadcast_event.mock_calls
        assert len(calls) == 2
        # First call is to check for new devices
        assert calls[0][1] == (KaraboEvent.SystemTopologyUpdate,
                               topo_update)
        # Second call is to clear with gone devices
        assert calls[1][1] == (KaraboEvent.ClearConfigurator,
                               {'devices': []})


def test_handle_instance_gone(mocker):
    network, topology = mocker.Mock(), mocker.Mock()
    with singletons(network=network, topology=topology):
        manager = Manager()

        devices = [('orphan', 'Parentless', ProxyStatus.OFFLINE)]
        topology.topology_update.return_value = (devices, [])

        calls = [
            mocker.call(KaraboEvent.SystemTopologyUpdate,
                        {'devices': devices, 'servers': []}),
            mocker.call(KaraboEvent.ClearConfigurator,
                        {'devices': ['orphan']})
        ]

        target = 'karabogui.singletons.manager.broadcast_event'
        broadcast_event = mocker.patch(target)
        changes = Hash("new", Hash(),
                       "update", Hash(),
                       "gone", Hash("device", "orphan"))
        manager.handle_topologyUpdate(changes)
        broadcast_event.assert_has_calls(calls)


def test_handle_instance_updated(gui_app, mocker):
    topology = mocker.Mock()
    mediator = Mediator()
    with singletons(topology=topology, mediator=mediator):
        topology.topology_update.return_value = [], []
        manager = Manager()
        changes = Hash("new", Hash(), "update", Hash(),
                       "gone", Hash())
        manager.handle_topologyUpdate(changes)
        topology.topology_update.assert_called_with(changes)

        target = 'karabogui.singletons.manager.broadcast_event'
        broadcast_event = mocker.patch(target)
        attr = Hash({'type': 'device',
                     'classId': 'BeckhoffMC2Beckhoff',
                     'serverId': 'test_server',
                     'host': 'localhost', 'status': 'ok', })

        update = Hash("device", Hash("MY_TEST_DEVICE", attr))

        changes = Hash("new", Hash(),
                       "update", update, "gone", Hash())
        manager.handle_topologyUpdate(changes)

        calls = broadcast_event.mock_calls
        assert broadcast_event.call_count == 3
        event, args = calls[0][1]
        assert event == KaraboEvent.SystemTopologyInstanceUpdate
        assert args == update


def test_handle_class_schema(mocker):
    network, topology = mocker.Mock(), mocker.Mock()
    with singletons(network=network, topology=topology):
        manager = Manager()

        schema = Schema()
        manager.handle_classSchema('server', 'ClassName', schema)

        topology.class_schema_updated.assert_called_with(
            'server', 'ClassName', schema)


def test_init_device_none(mocker):
    network, topology = mocker.Mock(), mocker.Mock()
    with singletons(network=network, topology=topology):
        # instantiate with wrong class def
        manager = Manager()
        topology.get_project_device_proxy.return_value = None

        cfg = Hash('init_prop', 42, 'ro_prop', -1)
        manager.initDevice(TEST_SERVER_ID, TEST_CLASS_ID, TEST_DEVICE_ID,
                           config=cfg)
        # No proxy means, not called!
        network.onInitDevice.assert_not_called()


def test_init_device_schema_evolution(mocker):
    network, topology = mocker.Mock(), mocker.Mock()
    with singletons(network=network, topology=topology):
        manager = Manager()
        schema = PrettyDevice.getClassSchema()
        topology.get_schema.return_value = schema
        proxy = _get_project_device_proxy(TEST_SERVER_ID, TEST_CLASS_ID,
                                          TEST_DEVICE_ID)

        topology.get_project_device_proxy.return_value = proxy

        cfg = Hash('init_prop', 42, 'ro_prop', -1, 'evolved', 43)
        manager.initDevice(TEST_SERVER_ID, TEST_CLASS_ID, TEST_DEVICE_ID,
                           config=cfg)
        topology.get_schema.assert_called_with(TEST_SERVER_ID,
                                               TEST_CLASS_ID)
        network.onInitDevice.assert_called_with(
            TEST_SERVER_ID, TEST_CLASS_ID, TEST_DEVICE_ID,
            # The configuration will be stripped of all the keys
            # that are read-only or not in the schema
            Hash('init_prop', 42))


def test_init_device_empty_config(mocker):
    network, topology = mocker.Mock(), mocker.Mock()
    with singletons(network=network, topology=topology):
        manager = Manager()
        schema = PrettyDevice.getClassSchema()
        topology.get_schema.return_value = schema
        proxy = _get_project_device_proxy(TEST_SERVER_ID, TEST_CLASS_ID,
                                          TEST_DEVICE_ID)
        topology.get_project_device_proxy.return_value = proxy

        manager.initDevice(TEST_SERVER_ID, TEST_CLASS_ID, TEST_DEVICE_ID)

        topology.get_schema.assert_called_with(TEST_SERVER_ID,
                                               TEST_CLASS_ID)
        network.onInitDevice.assert_called_with(
            TEST_SERVER_ID, TEST_CLASS_ID, TEST_DEVICE_ID,
            Hash())


def test_init_device_badschema(mocker):
    network, topology = mocker.Mock(), mocker.Mock()
    with singletons(network=network, topology=topology):
        manager = Manager()
        topology.get_schema.return_value = None
        proxy = _get_project_device_proxy(TEST_SERVER_ID, TEST_CLASS_ID,
                                          TEST_DEVICE_ID)
        topology.get_project_device_proxy.return_value = proxy

        target = 'karabogui.messagebox.show_error'
        msg_box = mocker.patch(target)
        cfg = Hash('init_prop', 42, 'ro_prop', -1)
        manager.initDevice(TEST_SERVER_ID, TEST_CLASS_ID,
                           TEST_DEVICE_ID, config=cfg)
        assert msg_box.call_count == 1


def test_project_db_handler(mocker):
    bad_result = Hash('value', 1, 'reason', 'error_msg')
    good_result = Hash('value', 1)
    target = 'karabogui.singletons.manager.messagebox'
    msg_box = mocker.patch(target)
    # this creates a function wrapped by the handler that returns
    # the input data
    request = Hash()
    handler = make_project_db_handler(False)
    handler('placeholder', False, request, bad_result, reason="error_msg")
    # error shown in case of bad result
    msg_box.show_error.assert_called_with('error_msg', details=None)

    handled_result = handler('placeholder', True, request,
                             good_result, reason="")
    assert handled_result.get('value') == 1

    # returning results in case `fall_through` is True
    handler = make_project_db_handler(True)
    handled_result = handler('placeholder', False, request, bad_result)
    assert handled_result.get('value') == 1

    handled_result = handler('placeholder', True, request, good_result)
    assert handled_result.get('value') == 1


def test_on_received_data(gui_app, mocker):
    manager = Manager()
    mocker.patch.object(manager, 'handle_requestGeneric')
    hsh = Hash('type', 'requestGeneric', 'value', 'terror')
    manager.onReceivedData(hsh)
    manager.handle_requestGeneric.assert_called_with(value='terror')


def test_on_server_connection_changed(mocker):
    network, topology = mocker.Mock(), mocker.Mock()
    with singletons(network=network, topology=topology):
        manager = Manager()
        target = 'karabogui.singletons.manager.broadcast_event'
        broadcast_event = mocker.patch(target)
        manager.onServerConnectionChanged(True)
        broadcast_event.assert_called_with(
            KaraboEvent.NetworkConnectStatus,
            {'status': True}
        )

        # Check that handlers are all called when network disconnects
        target = 'karabogui.singletons.manager.get_logger'

        logger = mocker.patch(target)
        manager._request_handlers = {'bob': 'super_request'}
        manager.onServerConnectionChanged(False)
        logger().error.assert_called_with(
            'Erased <b>1</b> pending generic requests due to '
            'client disconnect.')
        topology.clear.assert_called_with()
        # No further requests
        assert not manager._request_handlers


def test_handle_device_configuration(gui_app, mocker):
    topology = mocker.Mock()
    with singletons(topology=topology):
        manager = Manager()
        manager.handle_deviceConfiguration('Pauly', Hash('choochness', 11))
        topology.device_config_updated.assert_called_with(
            'Pauly', Hash('choochness', 11)
        )


def test_handle_project_list_items(gui_app, mocker):
    manager = Manager()
    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    h = Hash('success', True, 'items', ['schleem', 'plumbus'])
    manager.handle_projectListItems(True, Hash(), h)
    broadcast_event.assert_called_with(
        KaraboEvent.ProjectItemsList,
        {'items': ['schleem', 'plumbus']}
    )


def test_handle_project_list_domains(gui_app, mocker):
    manager = Manager()
    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    h = Hash('success', True, 'domains', ['WESTEROS'])
    manager.handle_projectListDomains(True, Hash(), h)
    broadcast_event.assert_called_with(
        KaraboEvent.ProjectDomainsList,
        {'items': ['WESTEROS']}
    )


def test_handle_project_load_items(gui_app, mocker):
    manager = Manager()
    h = Hash('items', ['load_me'], 'reason', 'we failed')

    broadcast = 'karabogui.singletons.manager.broadcast_event'
    messagebox = 'karabogui.singletons.manager.messagebox'
    broadcast_event = mocker.patch(broadcast)
    mocker.patch(messagebox)
    manager.handle_projectLoadItems(False, Hash(), h)
    broadcast_event.assert_called_with(
        KaraboEvent.ProjectItemsLoaded,
        {'success': False, 'items': ['load_me']}
    )


def test_handle_project_save_items(gui_app, mocker):
    manager = Manager()
    h = Hash('success', True, 'items', ['remember_this'])

    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    manager.handle_projectSaveItems(True, Hash(), h)
    broadcast_event.assert_called_with(
        KaraboEvent.ProjectItemsSaved,
        {'items': ['remember_this'], 'success': True}
    )


def test_handle_project_update_trashed(gui_app, mocker):
    manager = Manager()
    h = Hash('domain', "CONTROLS")

    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    manager.handle_projectUpdateTrashed(True, Hash(), h)
    broadcast_event.assert_called_with(
        KaraboEvent.ProjectTrashed,
        {'domain': 'CONTROLS'}
    )


def test_handle_projectUpdate(gui_app, mocker):
    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    manager = Manager()
    client = "my_client"
    uuids = ["123", "456"]
    h = Hash('success', True, 'info', {"client": client,
                                       "projects": uuids})
    manager.handle_projectUpdate(**h)
    broadcast_event.assert_called_with(
        KaraboEvent.ProjectUpdated, {'uuids': ['123', '456']})


def test_handle_set_log_reply(gui_app, mocker):
    manager = Manager()
    target = 'karabogui.messagebox.show_error'
    mb = mocker.patch(target)
    h = Hash("success", True)
    h["input"] = Hash("instanceId", "swerver",
                      "level", "DEBUG")
    manager.handle_setLogLevelReply(**h)
    mb.assert_not_called()
    h = Hash("success", False, "reason", "")
    h["input"] = Hash("instanceId", "swerver",
                      "level", "DEBUG")
    manager.handle_setLogLevelReply(**h)
    mb.assert_called_once()


def test_handle_network_data(gui_app, mocker):
    dev_proxy = mocker.Mock()
    prop_binding = mocker.Mock()
    topology = mocker.Mock()

    executeLater = 'karabogui.singletons.manager.executeLater'
    apply_tgt = 'karabogui.singletons.manager.apply_fast_data'
    later = mocker.patch(executeLater)
    apply_later = mocker.patch(apply_tgt)
    with singletons(topology=topology):
        manager = Manager()
        dev_proxy.get_property_binding.return_value = prop_binding
        topology.get_device.return_value = dev_proxy
        ts = Timestamp("2009-04-20T10:32:22 UTC")
        meta = Hash('timestamp', True)
        ts.toHashAttributes(meta)
        data = Hash('Position', 10)
        manager.handle_networkData('frankie:output', data, meta)
        assert 'frankie:output' in manager._big_data
        assert later.call_count == 1

        callback = later.mock_calls[0][1][0]
        callback()
        apply_later.assert_called_with(data, prop_binding.value.schema, ts)


def test_handle_init_reply(gui_app, mocker):
    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    manager = Manager()

    manager.handle_initReply('rick', True, 'wubbalubbadubdub')
    broadcast_event.assert_called_with(
        KaraboEvent.DeviceInitReply, mocker.ANY
    )
    event_data = broadcast_event.mock_calls[0][1][1]
    assert event_data['success']
    assert event_data['message'] == 'wubbalubbadubdub'
    assert isinstance(event_data['device'], DeviceProxy)

    mbox = mocker.patch('karabogui.singletons.manager.messagebox')
    info = {'deviceId': 'bob', 'message': 'mandatory key missing',
            'success': False}
    manager = Manager()
    manager.handle_initReply(**info)
    mbox.show_error.assert_called_with(
        'The instance <b>bob</b> could not be instantiated.. '
        '<br><br>The reason is:<br><i>mandatory key '
        'missing</i><br>', details=None)

    # details case
    info = {
        'deviceId': 'bob',
        'message': 'mandatory key missing\nDetails:\nhost missing',
        'success': False}
    manager.handle_initReply(**info)
    mbox.show_error.assert_called_with(
        'The instance <b>bob</b> could not be instantiated.. '
        '<br><br>The reason is:<br><i>mandatory key '
        'missing</i><br><br>Click '
        '"Show Details..." for more information.',
        details="host missing")


def test_handle_log_messages(gui_app, mocker):
    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    manager = Manager()
    messages = [Hash(
        "type", "error",
        "category", "XFEL/MOTOR/1",
        "message", "This is a test message",
        "traceback", "",
        "timestamp", Timestamp().toLocal())]
    manager.handle_log(messages)
    broadcast_event.assert_called_with(
        KaraboEvent.LogMessages, {"messages": messages})


def test_handle_broker_information(mocker):
    network = mocker.Mock()
    mediator = Mediator()
    with singletons(network=network, mediator=mediator):
        # New protocol
        manager = Manager()
        manager.handle_serverInformation(readOnly=True)
        network.set_server_information.assert_called_once_with(
            read_only=True)
        # Old protocol
        network.reset_mock()
        manager.handle_brokerInformation(readOnly=True)
        network.set_server_information.assert_called_once_with(
            read_only=True)


def test_handle_login_information(mocker):
    network = mocker.Mock()
    mediator = Mediator()
    path = 'karabogui.singletons.manager.broadcast_event'
    broad = mocker.patch(path)
    with singletons(network=network, mediator=mediator):
        manager = Manager()
        manager.handle_loginInformation(
            accessLevel=AccessLevel.EXPERT.value, username="karabo")
        broad.assert_called_with(KaraboEvent.LoginUserChanged, {})
        broad.reset_mock()
        # Try again downgrade
        manager.handle_loginInformation(
            accessLevel=AccessLevel.OBSERVER.value, username="karabo")
        broad.assert_called_with(KaraboEvent.LoginUserChanged, {})
        broad.reset_mock()
        # Read Only, but since we are observer, it is not changed
        manager.handle_loginInformation(
            accessLevel=AccessLevel.OBSERVER.value, username="karabo")
        assert krb_access.HIGHEST_ACCESS_LEVEL == AccessLevel.OBSERVER
        broad.assert_called_with(KaraboEvent.LoginUserChanged, {})
        broad.reset_mock()

        manager.handle_loginInformation(
            readOnly=True, username="karabo")
        assert krb_access.HIGHEST_ACCESS_LEVEL == AccessLevel.OBSERVER
        broad.assert_called_with(KaraboEvent.LoginUserChanged, {})
        broad.reset_mock()

        manager.handle_loginInformation(
            accessLevel=AccessLevel.EXPERT.value, username="karabo")
        broad.assert_called_with(KaraboEvent.LoginUserChanged, {})
        broad.reset_mock()
        assert krb_access.HIGHEST_ACCESS_LEVEL == AccessLevel.EXPERT


def test_handle_login_information_set_read_only(mocker):
    network = mocker.Mock()
    mediator = Mediator()
    # To avoid the influence of other tests, resetting SERVER_READ_ONLY
    krb_access.SERVER_READ_ONLY = None
    with singletons(network=network, mediator=mediator):
        manager = Manager()
        assert krb_access.SERVER_READ_ONLY is None
        manager.handle_loginInformation(
            accessLevel=AccessLevel.EXPERT.value, username="karabo",
            readOnly=False)
        # Make sure it is False and not None
        assert krb_access.SERVER_READ_ONLY is False

        # read-only state is changed only once.
        manager.handle_loginInformation(
            accessLevel=AccessLevel.OBSERVER.value, username="karabo",
            readOnly=True)
        assert not krb_access.SERVER_READ_ONLY

        # Observer set the read-only state
        krb_access.SERVER_READ_ONLY = None
        manager.handle_loginInformation(
            accessLevel=AccessLevel.OBSERVER.value, username="karabo",
            readOnly=True)
        assert krb_access.SERVER_READ_ONLY


def test_handle_property_history(gui_app, mocker):
    topology, device_proxy = mocker.Mock(), mocker.Mock()
    with singletons(topology=topology):
        topology.get_device.return_value = device_proxy
        manager = Manager()
        info = {'deviceId': 'bob', 'property': 'answer', 'data': [42],
                'success': True}
        manager.handle_propertyHistory(**info)
        expected_call = mocker.call.publish_historic_data('answer', [42])
        assert device_proxy.method_calls[0] == expected_call


def test_handle_notification(gui_app, mocker):
    m_path = 'karabogui.singletons.manager.messagebox'
    b_path = 'karabogui.singletons.manager.broadcast_event'
    l_path = 'karabogui.singletons.manager.get_logger'

    mbox = mocker.patch(m_path)
    broadcast = mocker.patch(b_path)
    info = {'message': 'hello'}
    manager = Manager()
    manager.handle_notification(**info)
    mbox.show_warning.assert_called_with('hello')

    mbox.reset_mock()
    info = {'nomessage': 'hello'}
    manager.handle_notification(**info)
    mbox.show_warning.assert_not_called()

    mbox.reset_mock()
    info = {'message': 'hello', 'contentType': 'banner'}
    manager.handle_notification(**info)
    mbox.show_warning.assert_not_called()
    broadcast.assert_called_with(
        KaraboEvent.ServerNotification, info)

    mbox.reset_mock()
    broadcast.reset_mock()
    logger = mocker.patch(l_path)
    info = {'message': 'hello', 'contentType': 'logger'}
    manager.handle_notification(**info)
    mbox.show_warning.assert_not_called()
    broadcast.assert_not_called()
    logger().log.assert_called_with(20, 'hello')

    asserts = [
        ('DEBUG', 10),
        ('INFO', 20),
        ('WARNING', 30),
        ('WARN', 30),
        ('ERROR', 40),
        ('FATAL', 50),
        ('CRITICAL', 50),
    ]
    for level, integer in asserts:
        logger().reset_mock()
        info = {'message': 'hello', 'contentType': 'logger', 'level': level}
        manager.handle_notification(**info)
        logger().log.assert_called_with(integer, 'hello')


def test_handle_executeReply(gui_app, mocker):
    target = 'karabogui.singletons.manager.messagebox'
    mbox = mocker.patch(target)
    manager = Manager()
    info = Hash(
        "success", False,
        "input", Hash("deviceId", "XFEL/MOTOR/2", "command", "move"),
        "reason", "timeout in gui server\nDetails:\nNot online")
    manager.handle_executeReply(**info)
    mbox.show_error.assert_called_with(
        'Execute slot <b>move</b> of device <b>XFEL/MOTOR/2</b> has '
        'encountered an error!<br><br>The reason is:'
        '<br><i>timeout in gui server</i><br><br>Click '
        '"Show Details..." for more information.',
        details="Not online")

    # No details test
    info = Hash(
        "success", False,
        "input", Hash("deviceId", "XFEL/MOTOR/2", "command", "move"),
        "reason", "timeout in gui server")
    manager.handle_executeReply(**info)
    mbox.show_error.assert_called_with(
        'Execute slot <b>move</b> of device <b>XFEL/MOTOR/2</b> has '
        'encountered an error!<br><br>The reason is:'
        '<br><i>timeout in gui server</i><br>',
        details=None)


def test_handle_deviceConfigurations(gui_app, mocker):
    topology = mocker.Mock()
    with singletons(topology=topology):
        manager = Manager()
        configs = {"XFEL/MOTOR/1": Hash("state", "MOVING"),
                   "XFEL/MOTOR/2": Hash("state", "OFF")}
        manager.handle_deviceConfigurations(configs)
        topology.device_config_updated.call_count == 2


def test_handle_configurationFromPast(gui_app, mocker):
    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    manager = Manager()
    config_time = Timestamp()
    config = Hash("archive", False)
    info = Hash(
        "success", True,
        "config", config,
        "reason", "",
        "deviceId", "XFEL/MOTOR/2",
        "configTimepoint", config_time,
        "configAtTimepoint", True,
        "preview", True,
        "time", config_time)

    manager.handle_configurationFromPast(**info)
    broadcast_event.assert_called_with(
        KaraboEvent.ShowConfigurationFromPast,
        {'deviceId': 'XFEL/MOTOR/2',
         'configuration': config,
         'time': config_time.toLocal(),
         'config_time': config_time.toLocal(),
         'preview': True, 'time_match': True})

    target = 'karabogui.singletons.manager.messagebox'
    mbox = mocker.patch(target)
    manager = Manager()
    timepoint = Timestamp()
    info = Hash(
        "success", False,
        "deviceId", "XFEL/MOTOR/2",
        "time", timepoint)
    reason = "Device not logged"
    info["reason"] = reason
    manager.handle_configurationFromPast(**info)
    mbox.show_error.assert_called_with(
        "The configuration of `XFEL/MOTOR/2` requested at time point "
        f"`{timepoint.toLocal()}` was not retrieved!<br><br>"
        f"The reason is:<br><i>{reason}</i>",
        details=None)


def test_handle_listInitConfigurations(gui_app, mocker):
    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    manager = Manager()
    items = ["a", "b", "c"]
    args = Hash("deviceId", "XFEL/CAM/1")
    info = Hash(
        "success", True,
        "reply", Hash("items", items),
        "request", Hash("args", args),
        "reason", "")

    manager.handle_listInitConfigurations(**info)
    broadcast_event.assert_called_with(
        KaraboEvent.ListConfigurationUpdated,
        {'items': items, 'deviceId': 'XFEL/CAM/1'})

    target = 'karabogui.singletons.manager.messagebox'
    mbox = mocker.patch(target)
    manager = Manager()
    info["success"] = False
    manager.handle_listInitConfigurations(**info)
    mbox.show_error.assert_called_with(
        'Requesting a list of configurations for XFEL/CAM/1 failed!',
        details='')


def test_handle_getInitConfiguration(gui_app, mocker):
    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    manager = Manager()
    item = Hash("name", "devName",
                "config", Hash("state", "ON"),
                "classId", "FooClass")

    args = Hash("deviceId", "XFEL/CAM/1")
    info = Hash(
        "success", True,
        "reply", Hash("item", item),
        "request", Hash("args", args, "preview", False),
        "reason", "")

    manager.handle_getInitConfiguration(**info)
    broadcast_event.assert_called_with(
        KaraboEvent.ShowInitConfiguration,
        {'configuration': Hash("state", "ON"), 'preview': False,
         'classId': 'FooClass',
         'name': 'devName', 'deviceId': 'XFEL/CAM/1'})

    target = 'karabogui.singletons.manager.messagebox'
    mbox = mocker.patch(target)
    manager = Manager()
    info["success"] = False
    manager.handle_getInitConfiguration(**info)
    mbox.show_error.assert_called_with(
        'Requesting a configuration for XFEL/CAM/1 failed!', details='')


def test_handle_saveInitConfiguration(gui_app, mocker):
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        item = Hash("name", "devName",
                    "config", Hash("state", "ON"))

        args = Hash("deviceIds", ["XFEL/CAM/1"])
        info = Hash(
            "success", True,
            "reply", Hash("item", item),
            "request", Hash("args", args, "preview", False,
                            "update", True),
            "reason", "")
        manager.handle_saveInitConfiguration(**info)
        network.onListInitConfigurations.assert_called_with(
            'XFEL/CAM/1')

    target = 'karabogui.singletons.manager.messagebox'
    mbox = mocker.patch(target)
    manager = Manager()
    info["success"] = False
    manager.handle_saveInitConfiguration(**info)
    mbox.show_error.assert_called_with(
        'Saving a configuration for XFEL/CAM/1 failed!', details='')


def test_handle_destinations(mocker):
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        broadcast = mocker.patch(
            "karabogui.singletons.manager.broadcast_event")
        reply = Hash("destinations", ["one", "two", "three"])
        manager.handle_listDestinations(success=True,
                                        request=Hash(), reply=reply)
        broadcast.assert_called_with(
            KaraboEvent.ActiveDestinations, ["one", "two", "three"])


def test_handle_saveLogBook(mocker):
    logger = mocker.patch("karabogui.singletons.manager.get_logger")
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        args = Hash("dataType", "image")
        h = Hash("args", args)
        manager.handle_saveLogBook(success=True, request=h, reply=h)
        message = "Posted the image to LogBook successfully"
        logger().info.assert_called_with(message)
        assert logger().info.call_count == 1


def test_handle_onBeginTemporarySession(mocker):
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        broadcast = mocker.patch(
            "karabogui.singletons.manager.broadcast_event")
        manager.handle_onBeginTemporarySession(success=False)
        assert broadcast.call_count == 0
        info = {"success": True, "accessLevel": 1}
        manager.handle_onBeginTemporarySession(**info)
        assert broadcast.call_count == 3
        first_call, second_call, third_call = broadcast.call_args_list
        args, _ = first_call
        assert args == (KaraboEvent.AccessLevelChanged, {})
        args, _ = second_call
        assert args == (KaraboEvent.LoginUserChanged, {})
        args, _ = third_call
        assert args == (KaraboEvent.UserSession, {})


def test_handle_onEndTemporarySession(mocker):
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        broadcast = mocker.patch(
            "karabogui.singletons.manager.broadcast_event")

        manager.handle_onEndTemporarySession(levelBeforeTemporarySession=2,
                                             loggedUserId="karabo")
        assert broadcast.call_count == 3
        first_call, second_call, third_call = broadcast.call_args_list
        args, _ = first_call
        assert args == (KaraboEvent.AccessLevelChanged, {})
        args, _ = second_call
        assert args == (KaraboEvent.LoginUserChanged, {})
        args, _ = third_call
        assert args == (KaraboEvent.UserSession, {})


def test_handle_onTemporarySessionExpired(mocker):
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        broadcast = mocker.patch(
            "karabogui.singletons.manager.broadcast_event")
        manager.handle_onTemporarySessionExpired(
            levelBeforeTemporarySession=2)
        assert broadcast.call_count == 2
        first_call, second_call = broadcast.call_args_list
        args, _ = first_call
        assert args == (KaraboEvent.LoginUserChanged, {})
        args, _ = second_call
        assert args == (KaraboEvent.UserSession, {})


def test_handle_onEndTemporarySessionNotice(mocker):
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        broadcast = mocker.patch(
            "karabogui.singletons.manager.broadcast_event")
        manager.handle_onEndTemporarySessionNotice()
        assert broadcast.call_count == 1


def test_temp_session(mocker):
    network = mocker.Mock()
    with singletons(network=network):
        manager = Manager()
        login_info = {"username": "abcd", "accessLevel": 2}
        mocker.patch("karabogui.singletons.manager.broadcast_event")
        manager.handle_loginInformation(**login_info)

        assert krb_access.GLOBAL_ACCESS_LEVEL == AccessLevel(2)
        assert krb_access.HIGHEST_ACCESS_LEVEL == AccessLevel(2)
        assert krb_access.TEMPORARY_SESSION_USER is None

        begin_session_info = {
            "username": "karabo", "accessLevel": 1, "success": True}
        manager.handle_onBeginTemporarySession(**begin_session_info)

        assert krb_access.GLOBAL_ACCESS_LEVEL == AccessLevel(1)
        assert krb_access.HIGHEST_ACCESS_LEVEL == AccessLevel(1)
        assert krb_access.TEMPORARY_SESSION_USER == "karabo"

        end_session_info = {"levelBeforeTemporarySession": 2}
        manager.handle_onEndTemporarySession(**end_session_info)

        assert krb_access.GLOBAL_ACCESS_LEVEL == AccessLevel(2)
        assert krb_access.HIGHEST_ACCESS_LEVEL == AccessLevel(2)
        assert krb_access.TEMPORARY_SESSION_USER is None

        temp_session_expired_info = {"levelBeforeTemporarySession": 1}
        manager.handle_onTemporarySessionExpired(**temp_session_expired_info)
        assert krb_access.GLOBAL_ACCESS_LEVEL == AccessLevel(1)
        assert krb_access.HIGHEST_ACCESS_LEVEL == AccessLevel(1)
        assert krb_access.TEMPORARY_SESSION_USER is None


def test_handle_session_info(mocker):
    target = 'karabogui.singletons.manager.broadcast_event'
    broadcast_event = mocker.patch(target)
    manager = Manager()
    info = Hash(
        "success", True, "reason", "",
        "type", "getGuiSessionInfo",
        "sessionStartTime", "2009-04-20T10:32:22 UTC",
        "sessionDuration", 120)

    manager.handle_getGuiSessionInfo(**info)
    data = {"sessionStartTime": "2009-04-20T10:32:22 UTC",
            "sessionDuration": 120}
    broadcast_event.assert_called_with(KaraboEvent.UserSessionInfo, data)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabo.common.scenemodel.api import TableElementModel
from karabo.native import Bool, Configurable, Hash, Int32, String, VectorHash
from karabogui.binding.config import apply_configuration
from karabogui.testing import get_class_property_proxy

from ..table import EditableTableElement


class Row(Configurable):
    foo = Bool(displayedName="Foo")
    bar = String(displayedName="Bar")
    cat = Int32(displayedName="Cat", options=[1, 2, 3])


class Object(Configurable):
    prop = VectorHash(rows=Row)


TABLE_HASH = Hash("prop", [Hash("foo", True, "bar", "hello", "cat", 1),
                           Hash("foo", False, "bar", "test", "cat", 3),
                           Hash("foo", False, "bar", "No", "cat", 2),
                           Hash("foo", True, "bar", "Jo", "cat", 1)])


@pytest.fixture
def table_setup(gui_app):
    proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
    model = TableElementModel()
    controller = EditableTableElement(proxy=proxy, model=model)
    controller.create(None)
    apply_configuration(TABLE_HASH, proxy.root_proxy.binding)
    yield controller
    # teardown
    controller.destroy()
    assert controller.widget is None


def assert_model_row(controller, row, first_column, second_column,
                     third_column, size=None):
    model = controller._item_model
    assert model.index(row, 0).data() == first_column
    assert model.index(row, 1).data() == second_column
    assert model.index(row, 2).data() == str(third_column)

    data = controller._item_model._data
    bool_cast = first_column == "True"
    assert data[row] == Hash("foo", bool_cast,
                             "bar", second_column,
                             "cat", third_column)
    if size is not None:
        assert data == size


def test_table_set_value(table_setup):
    controller = table_setup
    assert_model_row(controller, 0, "True", "hello", 1)
    assert_model_row(controller, 1, "False", "test", 3)
    assert_model_row(controller, 2, "False", "No", 2)
    assert_model_row(controller, 3, "True", "Jo", 1)


def test_table_empty_value(table_setup):
    controller = table_setup
    model = controller._item_model
    assert model._data != []
    model.clear_model()
    assert model._data == []


def test_table_edit_value(table_setup):
    controller = table_setup
    controller.set_read_only(False)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtWidgets import QAction, QInputDialog
from traits.api import Instance, WeakRef

from karabo.common.scenemodel.api import (
    VectorRollGraphModel, build_graph_config, restore_graph_config)
from karabo.native import Timestamp
from karabogui import icons
from karabogui.binding.api import NDArrayBinding, VectorNumberBinding
from karabogui.controllers.api import (
    BaseBindingController, get_array_data, register_binding_controller)
from karabogui.graph.common.api import AuxPlots, create_button
from karabogui.graph.image.api import (
    KaraboImagePlot, KaraboImageView, RollImage)

MAX_NUM_VECTORS = 500


@register_binding_controller(
    ui_name='VectorRoll Graph',
    klassname='VectorRollGraph',
    binding_type=(NDArrayBinding, VectorNumberBinding),
    priority=0,
    can_show_nothing=False)
class ArrayRollGraph(BaseBindingController):
    # Our VectorRollGraph Model
    model = Instance(VectorRollGraphModel, args=())

    _plot = WeakRef(KaraboImagePlot)
    _image = Instance(RollImage)
    _timestamp = Instance(Timestamp, args=())

    def create_widget(self, parent):
        widget = KaraboImageView(parent=parent)
        widget.stateChanged.connect(self._change_model)
        widget.add_colorbar()
        widget.add_picker()
        widget.add_roi()
        widget.add_aux(AuxPlots.ProfilePlot)

        # Finalize
        toolbar = widget.add_toolbar()

        _btn_reset = create_button(
            checkable=False,
            icon=icons.reset,
            tooltip="Reset the image",
            on_clicked=self._reset_image)
        toolbar.add_button(_btn_reset)

        widget.plotItem.set_aspect_ratio(0)
        widget.plotItem.vb.disableAutoRange()

        # Get a reference for our plotting
        self._plot = widget.plot()

        # Set the rolling image with stack!
        self._image = RollImage()
        self._image.stack = self.model.maxlen

        # QActions
        widget.add_axes_labels_dialog()

        size_action = QAction('Image Size', widget)
        size_action.triggered.connect(self._configure_maxlen)
        widget.addAction(size_action)

        # Restore the model information
        widget.restore(build_graph_config(self.model))

        return widget

    # -----------------------------------------------------------------------
    # Qt Slots

    def _reset_image(self):
        self._image.reset()
        self._plot.vb.setRange(yRange=(0, self._image.stack))

    def _change_model(self, content):
        self.model.trait_set(**restore_graph_config(content))

    def _configure_maxlen(self):
        maxlen, ok = QInputDialog.getInt(self.widget, 'Vector Stack',
                                         'Maxlen:', self.model.maxlen, 5,
                                         MAX_NUM_VECTORS)
        if ok:
            self._image.stack = maxlen
            self.model.maxlen = maxlen
            self._plot.vb.setRange(yRange=(0, maxlen))

    # -----------------------------------------------------------------------

    def value_update(self, proxy):
        value, timestamp = get_array_data(proxy)
        if value is None:
            return

        if timestamp != self._timestamp:
            # Get new timestamp reference!
            self._timestamp = timestamp

            self._image.add(value)
            if self._image.data is None:
                return

            if not self._plot.image_set:
                self._plot.vb.setRange(xRange=(0, len(value)),
                                       yRange=(0, self._image.stack))

            self._plot.setData(self._image.data, update=False)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os
from unittest.mock import patch

import numpy as np
from qtpy.QtCore import QPointF

from karabo.common.scenemodel.api import CurveType
from karabogui.graph.common.api import CrosshairROI, ROITool, safe_log10
from karabogui.testing import GuiTestCase

from ..base import KaraboPlotView

X_ARRAY = np.arange(10)
Y_ARRAY = X_ARRAY ** 2

FILENAME = "foo"
NPY = ".npy"
NPZ = ".npz"
NAMES = ["curve_1", "curve_2"]


class TestPlotViewExport(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.widget = KaraboPlotView()

    def tearDown(self):
        super().tearDown()
        self.widget.destroy()
        self.widget = None
        for ext in [NPY, NPZ]:
            self._delete_file(FILENAME + ext)

    def test_export_curve_item(self):
        # Test empty plot widget
        self._assert_npy_export(exported=False)

        # Test curves with no data
        curve_1 = self.widget.add_curve_item(name=NAMES[0])
        self._assert_npy_export(has_data=False)

        # Test curves with data
        curve_1.setData(X_ARRAY, Y_ARRAY)
        self._assert_npy_export()

        # # Test multiple curves
        curve_2 = self.widget.add_curve_item(name=NAMES[1])
        self._assert_npz_export(names=NAMES, has_data=[True, False])

        # # Test multiple curves
        curve_2.setData(X_ARRAY, Y_ARRAY)
        self._assert_npz_export(names=NAMES, has_data=[True, True])

    def test_export_histogram(self):
        curve = self.widget.add_curve_item(name=NAMES[0])

        # Test curves with data
        curve.setData(np.concatenate((X_ARRAY, [10])), Y_ARRAY,
                      stepMode="center")
        self._assert_npy_export()

    def test_export_scatter_item(self):
        # Test curve fill with no data
        item = self.widget.add_scatter_item()
        self._assert_npy_export(has_data=False)

        item.setData(X_ARRAY, Y_ARRAY)
        self._assert_npy_export()

    def _assert_npy_export(self, exported=True, has_data=True):
        filename = FILENAME + NPY
        self._assert_exported(filename, exported)
        if exported:
            self._assert_ndarray(np.load(filename), has_data)

    def _assert_npz_export(self, exported=True, names=(None, None),
                           has_data=(True, True)):
        filename = FILENAME + NPZ
        self._assert_exported(filename, exported)

        if exported:
            zipped = np.load(filename)
            assert isinstance(zipped, np.lib.npyio.NpzFile)
            assert set(names) == set(zipped.files)
            for index, name in enumerate(names):
                self._assert_ndarray(zipped[name], has_data[index])

    def _assert_exported(self, filename, exported):
        self._delete_file(filename)
        assert not os.path.isfile(filename)

        with patch('karabogui.util.getSaveFileName', return_value=filename):
            self.widget.export()

        assert os.path.isfile(filename) is exported

    def _assert_ndarray(self, data, has_data=True):
        x_array = X_ARRAY if has_data else []
        y_array = Y_ARRAY if has_data else []
        assert isinstance(data, np.ndarray)
        np.testing.assert_array_equal(data[0], x_array)
        np.testing.assert_array_equal(data[1], y_array)

    def _delete_file(self, filename):
        if os.path.isfile(filename):
            os.remove(filename)


DEFAULT_CONFIG = {"x_grid": False, "y_grid": False,
                  "x_log": False, "y_log": False,
                  "x_invert": False, "y_invert": False,
                  "x_label": '', "y_label": '',
                  "x_units": '', "y_units": '',
                  "x_autorange": True, "y_autorange": True,
                  "roi_items": [], "roi_tool": 0,
                  "background": "transparent", "title": ""}

orders = np.arange(-5, 5, dtype=np.float64)
LOG_ARRAY = 10 ** orders
RANGE_TOLERANCE = 0.12  # or 12% percent


class _BasePlotTest(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.widget = KaraboPlotView()
        self.widget.configuration.update(DEFAULT_CONFIG)
        self.widget.show()
        self.viewbox = self.widget.plotItem.vb
        self._plot = None

    def tearDown(self):
        super().tearDown()
        self.widget.close()
        self.widget.destroy()
        self.widget = None

    def restore(self, **kwargs):
        config = DEFAULT_CONFIG.copy()
        config.update(kwargs)
        self.widget.restore(config)

    def set_data(self, x, y):
        self._plot.setData(x, y)
        # Trigger auto range
        self.viewbox.updateAutoRange()

    def set_log_scale(self, x=None, y=None):
        if x is not None:
            self.widget.set_log_x(x)
            self.widget.configuration['x_log'] = x
        if y is not None:
            self.widget.set_log_y(y)
            self.widget.configuration['y_log'] = y
        self.viewbox.updateAutoRange()

    def set_range_x(self, low, high):
        expected = {"x_min": low, "x_max": high, "x_autorange": False}
        with patch("karabogui.graph.plots.base.RangeDialog.get",
                   return_value=(expected, True)):
            self.widget.configure_range_x()

    def set_range_y(self, low, high):
        expected = {"y_min": low, "y_max": high, "y_autorange": False}
        with patch("karabogui.graph.plots.base.RangeDialog.get",
                   return_value=(expected, True)):
            self.widget.configure_range_y()

    def assert_data(self, x_expected, y_expected, log_x=False, log_y=False):
        # Check getData values
        x_data, y_data = self._plot.getData()
        if log_x:
            x_expected = safe_log10(x_expected)
            np.testing.assert_almost_equal(x_data, x_expected)
        else:
            np.testing.assert_array_equal(x_data, x_expected)
        if log_y:
            y_expected = safe_log10(y_expected)
            np.testing.assert_almost_equal(y_data, y_expected)
        else:
            np.testing.assert_array_equal(y_data, y_expected)

    def assert_range(self, x_expected, y_expected, log_x=False, log_y=False):
        self.assert_range_x(x_expected, log_x)
        self.assert_range_y(y_expected, log_y)

    def assert_range_x(self, x_expected, log_x):
        assert_x_range = (self.assert_log_range if log_x
                          else self.assert_linear_range)
        assert_x_range(actual=self.widget.get_view_range_x(),
                       expected=x_expected)

    def assert_range_y(self, y_expected, log_y):
        assert_y_range = (self.assert_log_range if log_y
                          else self.assert_linear_range)
        assert_y_range(actual=self.widget.get_view_range_y(),
                       expected=y_expected)

    def assert_log_range(self, actual, expected):
        """The tolerance is by its order with the deviation of
           12% the of range """
        actual = np.ceil(np.log10(actual))
        # Calculate expected range
        log_expected = safe_log10(expected)
        expected = np.nanmin(log_expected), np.nanmax(log_expected)
        # Calculate tolerance, expect a range of at least 1 order.
        tol = int((expected[-1] - expected[0]) * 0.12) or 1
        np.testing.assert_allclose(actual, expected, atol=tol)

    def assert_linear_range(self, actual, expected):
        tol = (expected[-1] - expected[0]) * RANGE_TOLERANCE
        np.testing.assert_allclose(actual, [expected[0], expected[-1]],
                                   atol=tol)


class TestCurveItem(_BasePlotTest):

    def setUp(self):
        super().setUp()
        self._plot = self.widget.add_curve_item()

    def test_basics(self):
        self.set_data(x=X_ARRAY, y=LOG_ARRAY)
        self.assert_data(x_expected=X_ARRAY, y_expected=LOG_ARRAY)
        self.assert_range(x_expected=X_ARRAY, y_expected=LOG_ARRAY)

    def test_log_scale(self):
        self.set_data(x=X_ARRAY, y=LOG_ARRAY)
        self.set_log_scale(y=True)
        self.assert_data(x_expected=X_ARRAY, y_expected=LOG_ARRAY, log_y=True)
        # Check if the range doesn't change with log y scale
        self.assert_range(x_expected=X_ARRAY, y_expected=LOG_ARRAY,
                          log_y=True)

    def test_ranges(self):
        self.set_data(x=X_ARRAY, y=LOG_ARRAY)
        self.set_log_scale(y=True)

        # Set range
        low, high = 1e-9, 1e9
        self.set_range_y(low, high)
        self.assert_range_y(y_expected=(low, high), log_y=True)

        # Set problematic range (with zeros)
        low, high = 0, 1e4
        self.set_range_y(low, high)
        corrected_low = 10 ** safe_log10(low)  # 1e-25
        self.assert_range_y(y_expected=(corrected_low, high), log_y=True)

    def test_legend_text(self):
        legend = self.widget.add_legend()
        plot = self.widget.add_curve_item(name="test_plot")
        label = legend.getLabel(plot)
        assert label.text == "test_plot"

        self.widget.update_legend_text(plot, "new_legend")
        label = legend.getLabel(plot)
        assert label.text == "new_legend"

    def test_apply_options(self):
        plot_1_color = self._plot.opts["pen"].color().name()
        legend = self.widget.add_legend()
        plot = self.widget.add_curve_item(name="test_plot")
        options = {plot: {"pen_color": "#ff7f00",
                          "curve_type": CurveType.Curve,
                          "name": "new_legend"}}
        self.widget.apply_curve_options(options)
        label = legend.getLabel(plot)
        assert label.text == "new_legend"
        plot.opts["pen"].color().name() == "#ff7f00"

        # When data points are shown, very they are updated too.
        self.widget._show_symbols = True
        self.widget.apply_curve_options(options)
        assert plot.opts["symbolPen"].color().name() == "#ff7f00"
        assert plot.opts["symbolBrush"].color().name() == "#ff7f00"

        # The other curve should have no change.
        assert legend.getLabel(self._plot) is None
        assert self._plot.opts["pen"].color().name() == plot_1_color


class TestBarItem(_BasePlotTest):

    def setUp(self):
        super().setUp()
        self._plot = self.widget.add_bar_item()
        self._plot.set_width(0.1)

    def test_basics(self):
        self.set_data(x=X_ARRAY, y=LOG_ARRAY)
        self._assert_bar(x_expected=X_ARRAY, y_expected=LOG_ARRAY)
        self.assert_data(x_expected=X_ARRAY, y_expected=LOG_ARRAY)
        self.assert_range(x_expected=X_ARRAY, y_expected=LOG_ARRAY)

    def test_log_scale(self):
        self.set_data(x=X_ARRAY, y=LOG_ARRAY)
        self.set_log_scale(y=True)
        self._assert_bar(x_expected=X_ARRAY, y_expected=LOG_ARRAY)
        self.assert_data(x_expected=X_ARRAY, y_expected=LOG_ARRAY, log_y=True)
        # Check if the range doesn't change with log y scale
        self.assert_range(x_expected=X_ARRAY, y_expected=LOG_ARRAY,
                          log_y=True)

    def test_problematic_log_scale(self):
        y_array = np.array([0, 0.001, 0, -0.001])
        x_array = np.arange(y_array.size)
        self.set_data(x=x_array, y=y_array)
        self.set_log_scale(y=True)
        self._assert_bar(x_expected=x_array, y_expected=y_array)
        self.assert_data(x_expected=x_array, y_expected=y_array, log_y=True)
        # Check if the range doesn't change with log y scale
        self.assert_range(x_expected=x_array, y_expected=y_array,
                          log_y=True)

    def _assert_bar(self, x_expected, y_expected):
        # Check bar attributes
        np.testing.assert_array_equal(self._plot.opts.get('x'),
                                      x_expected)
        np.testing.assert_array_equal(self._plot.opts.get('height'),
                                      y_expected)


class TestPlotViewRestore(_BasePlotTest):

    def test_restore_roi(self):
        controller = self.widget.add_roi()
        toolbar = self.widget.add_toolbar()

        # Restore ROI items
        roi_tool = int(ROITool.Crosshair)
        roi_items = [{"roi_type": roi_tool, "x": 15, "y": 15, "name": "ROI"}]
        self.restore(roi_items=roi_items, roi_tool=roi_tool)

        # Check if only one ROI type is present and current
        assert controller.current_tool == roi_tool
        assert len(controller.roi_items) == 1

        # Check if only one ROI item is present
        crosshairs = controller.roi_items[roi_tool]
        assert len(crosshairs) == 1

        # Check ROI item and coords
        crosshair = crosshairs[0]
        assert isinstance(crosshair, CrosshairROI)
        x, y = crosshair.coords
        assert x == 15
        assert y == 15
        assert crosshair.center == QPointF(15.000000, 15.000000)

        # Check if tool button is checked
        assert toolbar.buttons[ROITool(roi_tool).name].isChecked()

    def test_restore_manual_ranges(self):
        # Set range with with log
        low, high = 1e-9, 1e9
        self.restore(y_min=low, y_max=high, y_autorange=False, y_log=True)
        self.assert_range_y(y_expected=(low, high), log_y=True)

        # Set some data and check if range will not change
        self._plot = self.widget.add_curve_item()
        self.set_data(x=X_ARRAY, y=LOG_ARRAY)
        self.assert_range_y(y_expected=(low, high), log_y=True)

    def test_axis_visibility(self):
        bottom_axes = self.widget.plotItem.getAxis("bottom")
        left_axes = self.widget.plotItem.getAxis("left")
        x_label = bottom_axes.label
        y_label = left_axes.label
        assert x_label.isVisible() is False
        assert y_label.isVisible() is False

        self.restore(x_label="Axis", y_label="Axis")
        assert x_label.isVisible() is True
        assert y_label.isVisible() is True

        self.restore(x_label="", y_label="")
        assert x_label.isVisible() is False
        assert y_label.isVisible() is False

        self.restore(x_units="mm", y_units="mm")
        assert x_label.isVisible() is True
        assert y_label.isVisible() is True

        self.restore(x_units="", y_units="")
        assert x_label.isVisible() is False
        assert y_label.isVisible() is False

    def test_title_background(self):
        toolbar = self.widget.add_toolbar()
        self.restore(background="white")
        brush = self.widget.graph_view.backgroundBrush()
        assert brush.color().getRgb() == (255, 255, 255, 255)
        self.restore(background="red")
        brush = self.widget.graph_view.backgroundBrush()
        assert brush.color().getRgb() == (255, 0, 0, 255)
        assert self.widget._toolbar is not None

        tb = toolbar.widget
        sheet = tb.styleSheet()
        assert "background-color: rgba(255, 0, 0, 255)" in sheet



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 9, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QCompleter, QDialog, QDialogButtonBox

from karabogui.topology.api import getTopology
from karabogui.util import InputValidator

from .utils import get_dialog_ui


class ServerHandleDialog(QDialog):
    def __init__(self, model=None, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui("server_handle.ui"), self)

        servers = self.get_available_servers()
        self.ui_server_id.addItems(servers)
        completer = QCompleter(servers, parent=self)
        completer.setCaseSensitivity(False)
        completer.setCompletionMode(QCompleter.PopupCompletion)
        completer.setFilterMode(Qt.MatchContains)
        self.ui_server_id.setCompleter(completer)

        if model is None:
            title = "Add server"
        else:
            title = "Edit server"
            index = self.ui_server_id.findText(model.server_id)
            if index < 0:
                server_edit = self.ui_server_id.lineEdit()
                server_edit.setText(model.server_id)
            else:
                self.ui_server_id.setCurrentIndex(index)

        self.setWindowTitle(title)

        validator = InputValidator(object_type="server", parent=self)
        self.ui_server_id.setValidator(validator)

        self.ui_server_id.currentIndexChanged.connect(self._update_button_box)
        self.ui_server_id.editTextChanged.connect(self._update_button_box)
        self._update_button_box()

    def get_available_servers(self):
        """ Get all available servers of the `systemTopology`"""
        servers = list(getTopology()["server"])
        return sorted(servers)

    @Slot()
    def _update_button_box(self):
        """Only enable Ok button, if title and configuration is set
        """
        enabled = len(self.ui_server_id.currentText())
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    @property
    def server_id(self):
        return self.ui_server_id.currentText()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 30, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import json
from abc import abstractmethod

from qtpy.QtCore import QPoint
from qtpy.QtWidgets import QBoxLayout
from traits.api import ABCHasStrictTraits, Undefined

from karabo.common.enums import Capabilities
from karabo.common.scenemodel.api import (
    BoxLayoutModel, DeviceSceneLinkModel, LabelModel, SceneLinkModel)
from karabo.common.scenemodel.const import SceneTargetWindow
from karabo.native import AccessMode
from karabogui import messagebox
from karabogui.binding.api import ImageBinding, SlotBinding
from karabogui.controllers.api import (
    get_class_const_trait, get_compatible_controllers, get_scene_model_class)
from karabogui.fonts import get_font_metrics
from karabogui.itemtypes import NavigationItemTypes, ProjectItemTypes
from karabogui.sceneview.utils import round_down_to_grid
from karabogui.sceneview.widget.utils import get_proxy
from karabogui.singletons.api import get_topology

_STACKED_WIDGET_OFFSET = 30
_NO_LABEL_BINDINGS = (ImageBinding, SlotBinding)
_LINK_MARGIN = 10
_LINK_SIZE_HIT = 30

_DEVICE_TYPES = (NavigationItemTypes.DEVICE, ProjectItemTypes.DEVICE)


class SceneDnDHandler(ABCHasStrictTraits):
    @abstractmethod
    def can_handle(self, event):
        """Check whether the drag event can be handled."""

    @abstractmethod
    def handle(self, scene_view, event):
        """Handle the drop event."""


class ConfigurationDropHandler(SceneDnDHandler):
    """Scene D&D handler for drops originating from the configuration view.
    """

    def can_handle(self, event):
        source_type = event.mimeData().data('source_type')
        if source_type == 'ParameterTreeWidget':
            return True
        return False

    def handle(self, scene_view, event):
        mime_data = event.mimeData()
        items = []
        if mime_data.data('tree_items'):
            items_data = mime_data.data('tree_items').data()
            items = json.loads(items_data.decode())

        # Create the proxies first
        proxies = [get_proxy(*item['key'].split('.', 1)) for item in items]

        # Handle the case when dropped on an existing scene widget
        pos = event.pos()
        widget = scene_view.widget_at_position(pos)
        if widget is not None and widget.add_proxies(proxies):
            return

        # Handle the case when dropped as new scene widgets
        models = []
        for item, proxy in zip(items, proxies):
            if proxy.binding is None:
                continue

            model = self._create_model_from_parameter_item(item, proxy, pos)
            models.append(model)
            pos += QPoint(0, _STACKED_WIDGET_OFFSET)
        scene_view.add_models(*models, initialize=True)

    def _create_model_from_parameter_item(self, item, proxy, pos):
        """Create the scene models for a single item
        """

        def _create_model(klass, key):
            """Create a single widget from a class `klass` with a `key`"""
            model_klass = get_scene_model_class(klass)
            model = model_klass(keys=[key])
            model.x = round_down_to_grid(pos.x())
            model.y = round_down_to_grid(pos.y())
            if hasattr(model, 'klass'):
                model.klass = get_class_const_trait(klass, '_klassname')
            return model

        if isinstance(proxy.binding, _NO_LABEL_BINDINGS):
            # Note: Some bindings do not require a label widget and will
            # only appear with one widget
            klasses = get_compatible_controllers(proxy.binding)
            if klasses:
                return _create_model(klasses[0], proxy.key)

        # Horizonal layout
        layout_model = BoxLayoutModel(direction=QBoxLayout.LeftToRight,
                                      x=round_down_to_grid(pos.x()),
                                      y=round_down_to_grid(pos.y()))
        # Add label to layout model
        label_model = LabelModel(text=item['label'], foreground='#000000')
        layout_model.children.append(label_model)

        def insert_model_layout(klass, proxy, layout_model):
            model_klass = get_scene_model_class(klass)
            model = model_klass(keys=[proxy.key])
            klass.initialize_model(proxy, model)
            if hasattr(model, 'klass'):
                model.klass = get_class_const_trait(klass, '_klassname')
            layout_model.children.append(model)

        # Add the display and editable widgets, as needed
        klasses = get_compatible_controllers(proxy.binding)
        if klasses:
            insert_model_layout(klasses[0], proxy, layout_model)

        # Only editable if RECONFIGURABLE
        if proxy.binding.accessMode is AccessMode.RECONFIGURABLE:
            klasses = get_compatible_controllers(proxy.binding, can_edit=True)
            if klasses:
                insert_model_layout(klasses[0], proxy, layout_model)

        return layout_model


class ProjectDropHandler(SceneDnDHandler):
    """Scene D&D handler for drops originating from the project view"""

    def can_handle(self, event):
        # We can handle a drop if it contains at least one item
        items = self._extract_items(event.mimeData())
        return (len(items) > 0 and
                items[0].get('type') == ProjectItemTypes.SCENE)

    def handle(self, scene_view, event):
        dropped_items = self._extract_items(event.mimeData())
        if len(dropped_items) == 0:
            return

        item = dropped_items[0]
        if item.get('type') == ProjectItemTypes.SCENE:
            uuid = item.get('uuid')
            simple_name = item.get('simple_name')
            target = f'{simple_name}:{uuid}'
            position = event.pos()

            fm = get_font_metrics()
            width = max(fm.width(simple_name) + _LINK_MARGIN, _LINK_SIZE_HIT)
            height = max(fm.height() + _LINK_MARGIN, _LINK_SIZE_HIT)

            model = SceneLinkModel(target=target,
                                   text=simple_name,
                                   target_window=SceneTargetWindow.Dialog,
                                   x=position.x(), y=position.y(),
                                   width=width, height=height)
            scene_view.add_models(model, initialize=True)

    def _extract_items(self, mime_data):
        known_types = (ProjectItemTypes.DEVICE, ProjectItemTypes.SCENE)
        items_data = mime_data.data('treeItems').data()
        if items_data:
            items = json.loads(items_data.decode())
            return [it for it in items if it['type'] in known_types]
        return []


class NavigationDropHandler(SceneDnDHandler):
    """Scene D&D handler for drops originating from the project view"""

    def can_handle(self, event):
        # We can handle a drop if it contains at least one item
        items = self._extract_items(event.mimeData())
        if not len(items):
            return False

        from_device = items[0].get('type') in _DEVICE_TYPES
        capa = Capabilities.PROVIDES_SCENES
        has_scene = items[0].get('capabilities') & capa == capa
        return from_device and has_scene

    def handle(self, scene_view, event):
        dropped_items = self._extract_items(event.mimeData())
        if len(dropped_items) == 0:
            return

        item = dropped_items[0]
        if item.get('type') in _DEVICE_TYPES:
            device_id = item['deviceId']

            def attach_device_link(scene_name):
                """Attach a device scene link"""
                nonlocal event, device_id, scene_view

                fm = get_font_metrics()
                width = max(fm.width(device_id) + _LINK_MARGIN, _LINK_SIZE_HIT)
                height = max(fm.height() + _LINK_MARGIN, _LINK_SIZE_HIT)

                position = event.pos()
                model = DeviceSceneLinkModel(
                    text=device_id,
                    keys=[f"{device_id}.availableScenes"],
                    target=scene_name,
                    target_window=SceneTargetWindow.Dialog,
                    x=position.x(), y=position.y(),
                    width=width, height=height)
                scene_view.add_models(model, initialize=True)

            def _config_handler():
                """Act on the arrival of the configuration"""
                proxy.on_trait_change(_config_handler, 'config_update',
                                      remove=True)
                scenes = proxy.binding.value.availableScenes.value
                if scenes is Undefined or not len(scenes):
                    messagebox.show_warning(
                        "The device <b>{}</b> does not specify a scene "
                        "name!".format(device_id))
                else:
                    scene_name = scenes[0]
                    attach_device_link(scene_name)

            def _schema_handler():
                """Act on the arrival of the schema"""
                proxy.on_trait_change(_schema_handler, 'schema_update',
                                      remove=True)
                scenes = proxy.binding.value.availableScenes.value
                if scenes is Undefined:
                    proxy.on_trait_change(_config_handler, 'config_update')
                elif not len(scenes):
                    messagebox.show_warning(
                        "The device <b>{}</b> does not specify a scene "
                        "name!".format(device_id))
                else:
                    scene_name = scenes[0]
                    attach_device_link(scene_name)

            proxy = get_topology().get_device(device_id)
            if not len(proxy.binding.value):
                # We completely miss our schema and wait for it.
                proxy.on_trait_change(_schema_handler, 'schema_update')
            elif proxy.binding.value.availableScenes.value is Undefined:
                # The configuration did not yet arrive and we cannot get
                # a scene name from the availableScenes.
                proxy.on_trait_change(_config_handler, 'config_update')
            else:
                scenes = proxy.binding.value.availableScenes.value
                if not len(scenes):
                    messagebox.show_warning(
                        "The device <b>{}</b> does not specify a scene "
                        "name!".format(device_id))
                else:
                    scene_name = scenes[0]
                    attach_device_link(scene_name)

    def _extract_items(self, mime_data):
        items_data = mime_data.data('treeItems').data()
        if items_data:
            items = json.loads(items_data.decode())
            return [it for it in items if it['type'] in _DEVICE_TYPES]
        return []



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on October 4, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import numpy as np

from karabo.native import Encoding
from karabogui.binding.api import (
    REFERENCE_TYPENUM_TO_DTYPE, get_binding_array_value)

try:
    import cv2

    has_open_cv = True
except ImportError:
    has_open_cv = False

# Maps image physical dimension to ndarray dimension
DIMENSIONS = {
    'X': 1,
    'Y': 0,
    'Z': 2
}


def _extract_npy_data(image_node):
    """Internal method to extract npy data from an image node"""
    pixels = image_node.pixels.value
    if pixels.data.value is None:
        return None
    if len(pixels.data.value) == 0:
        return None
    arr_type = REFERENCE_TYPENUM_TO_DTYPE[pixels.type.value]
    npy = np.frombuffer(pixels.data.value, dtype=arr_type)
    return npy


def get_image_data(image_node, dimX, dimY, dimZ):
    """ Calculate a numpy array from the given ``image_node`` depending on the
    given dimensions and return it to use for image display.
    In case no data is included, a ``NoneType`` is returned.
    """
    npy = _extract_npy_data(image_node)
    if npy is None:
        return

    if dimZ:
        try:
            npy.shape = dimY, dimX, dimZ
        except ValueError:
            msg = 'Image has improper shape ({}, {}, {}) for size {}'.format(
                dimX, dimY, dimZ, len(npy))
            raise RuntimeError(msg)
    else:
        try:
            npy.shape = dimY, dimX
        except ValueError:
            msg = 'Image has improper shape ({}, {}, {}) for size {}'.format(
                dimX, dimY, dimZ, len(npy))
            raise RuntimeError(msg)

    return npy


def get_jpeg_data(image_node, dimZ):
    """ Calculate a numpy array from the given ``image_node`` depending on the
    given dimensions for a jpeg encoded stream
    """
    if not has_open_cv:
        return None

    npy = _extract_npy_data(image_node)
    if npy is None:
        return

    if dimZ == 3:
        npy = cv2.imdecode(npy, cv2.IMREAD_COLOR)
        # Default decode is BGR ...
        npy = cv2.cvtColor(npy, cv2.COLOR_BGR2RGB)
    else:
        npy = cv2.imdecode(npy, cv2.IMREAD_GRAYSCALE)

    return npy


def get_dimensions_and_encoding(image_node):
    """ The dimensions and the encoding of the given ``image_node`` are
    returned."""
    dimX = None
    dimY = None
    dimZ = None
    encoding = image_node.encoding.value

    if len(image_node.dims.value) == 2:
        # Shape
        dimX = image_node.dims.value[DIMENSIONS['X']]
        dimY = image_node.dims.value[DIMENSIONS['Y']]
        if encoding == Encoding.UNDEFINED:
            encoding = Encoding.GRAY  # assume it's gray
    elif len(image_node.dims.value) == 3:
        # Shape
        dimX = image_node.dims.value[DIMENSIONS['X']]
        dimY = image_node.dims.value[DIMENSIONS['Y']]
        dimZ = image_node.dims.value[DIMENSIONS['Z']]
        if encoding == Encoding.UNDEFINED:
            if dimZ == 3:
                encoding = Encoding.RGB  # assume it's RGB
            elif dimZ == 4:
                encoding = Encoding.RGBA  # assume it's RGBA
            else:
                encoding = Encoding.GRAY  # assume it's a stack of GRAY

    return dimX, dimY, dimZ, encoding


def get_array_data(proxy, default=None):
    """Retrieve the array and timestamp data from a property proxy belonging
    to an array binding

    :param default: default value to be returned if no value is available

    This function checks for `Undefined` on the proxy value and `None` data.
    If not data is available the `default` is returned with actual timestamp.

    :returns: data, timestamp
    """
    binding = proxy.binding
    return get_binding_array_value(binding, default)



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on February 10, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtWidgets import QAction, QInputDialog
from traits.api import Instance, on_trait_change

from karabo.common.scenemodel.api import (
    DoubleLineEditModel, EditableRegexModel, HexadecimalModel,
    IntLineEditModel)
from karabogui.binding.api import (
    FloatBinding, IntBinding, StringBinding, get_min_max)
from karabogui.controllers.api import (
    BaseLineEditController, get_regex, has_regex, register_binding_controller)
from karabogui.validators import (
    HexValidator, IntValidator, NumberValidator, RegexValidator)

MAX_FLOATING_PRECISION = 12


@register_binding_controller(ui_name="Float Field", can_edit=True,
                             klassname="DoubleLineEdit",
                             binding_type=FloatBinding, priority=10)
class DoubleLineEdit(BaseLineEditController):
    model = Instance(DoubleLineEditModel, args=())

    def create_widget(self, parent):
        widget = super().create_widget(parent)
        decimal_action = QAction("Change number of decimals", widget)
        decimal_action.triggered.connect(self._pick_decimals)
        widget.addAction(decimal_action)
        return widget

    # ----------------------------------------------------------------------
    # Abstract Interface

    def create_validator(self):
        """Reimplemented function of `BaseLineEditController`"""
        return NumberValidator(self.model.decimals)

    def binding_validator(self, proxy):
        """Reimplemented method of `BaseLineEditController`"""
        low, high = get_min_max(proxy.binding)
        self.validator.setBottom(low)
        self.validator.setTop(high)

    def toString(self, value):
        """Reimplemented method of `BaseLineEditController`"""
        format_str = ("{}" if self.model.decimals == -1
                      else f"{{:.{self.model.decimals}f}}")
        return format_str.format(float(str(value)))

    def fromString(self, value):
        """Reimplemented method of `BaseLineEditController`"""
        return float(value)

    # ----------------------------------------------------------------------

    @on_trait_change("model.decimals", post_init=True)
    def _decimals_update(self):
        self.value_update(self.proxy)
        self.validator.decimals = self.model.decimals

    def _pick_decimals(self, checked):
        num_decimals, ok = QInputDialog.getInt(
            self.widget, "Decimal", "Floating point precision:",
            self.model.decimals, -1, MAX_FLOATING_PRECISION)
        if ok:
            self.model.decimals = num_decimals


@register_binding_controller(ui_name="Integer Field", can_edit=True,
                             klassname="IntLineEdit", binding_type=IntBinding,
                             priority=10)
class IntLineEdit(BaseLineEditController):
    model = Instance(IntLineEditModel, args=())

    # ----------------------------------------------------------------------
    # Abstract Interface

    def create_validator(self):
        """Reimplemented method of `BaseLineEditController`"""
        return IntValidator()

    def binding_validator(self, proxy):
        """Reimplemented method of `BaseLineEditController`"""
        low, high = get_min_max(proxy.binding)
        self.validator.setBottom(low)
        self.validator.setTop(high)

    def fromString(self, value):
        """Reimplemented method of `BaseLineEditController`"""
        return int(value)


@register_binding_controller(ui_name="Hexadecimal", can_edit=True,
                             klassname="Hexadecimal", binding_type=IntBinding)
class Hexadecimal(BaseLineEditController):
    model = Instance(HexadecimalModel, args=())

    # ----------------------------------------------------------------------
    # Abstract Interface

    def create_validator(self):
        """Reimplemented method of `BaseLineEditController`"""
        return HexValidator()

    def binding_validator(self, proxy):
        """Reimplemented method of `BaseLineEditController`"""
        low, high = get_min_max(proxy.binding)
        self.validator.setBottom(low)
        self.validator.setTop(high)

    def toString(self, value):
        """Reimplemented method of `BaseLineEditController`"""
        return f"{value:x}"

    def fromString(self, value):
        """Reimplemented method of `BaseLineEditController`"""
        return int(value, base=16)


@register_binding_controller(ui_name="Regex Field", can_edit=True,
                             is_compatible=has_regex,
                             klassname="RegexEdit", binding_type=StringBinding,
                             priority=90)
class EditRegex(BaseLineEditController):
    model = Instance(EditableRegexModel, args=())

    # ----------------------------------------------------------------------
    # Abstract Interface

    def create_validator(self):
        """Reimplemented method of `BaseLineEditController`"""
        return RegexValidator()

    def binding_validator(self, proxy):
        """Reimplemented method of `BaseLineEditController`"""
        self.validator.setRegex(get_regex(proxy.binding, ""))



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on November 8, 2021
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from enum import IntEnum

from qtpy.QtCore import QPoint, QRect, Qt, Signal
from qtpy.QtGui import QBrush, QPainter, QPalette, QPen
from qtpy.QtWidgets import QSlider, QStyle, QStyleOptionSlider

INT32_MAX = 2147483647
INT32_MIN = -2147483648


class SliderHandle(IntEnum):
    """Enum to indicate the active handle of a RangeSlider"""
    LOW = 0
    HIGH = 1
    SCROLL_BAR = -1


def integer(func):
    """Decorate a function to ensure integer input of the parameters"""

    def wrapper(self, *args):
        args = [int(v) for v in args]
        return func(self, *args)

    return wrapper


class RangeSlider(QSlider):
    """The RangeSlider that can have low and high handle

    <|--Low---High-------|>

    This is an adapted and simplified `RangeSlider` covering two handles
    and the enabled and disabled version.
    """
    sliderMoved = Signal(int, int)  # Reimplemented signal

    def __init__(self, orientation=Qt.Horizontal, parent=None):
        super().__init__(orientation, parent=parent)

        self.low_handle_pos = self.minimum()
        self.high_handle_pos = self.maximum()

        self.pressedControl = QStyle.SC_None
        self.clickOffset = 0
        self.pressedHandle = None
        self.setTickPosition(self.NoTicks)
        self.setFocusPolicy(Qt.ClickFocus)

    def initStyleOption(self, option):
        super().initStyleOption(option)
        option.subControls = QStyle.SC_None
        option.activeSubControls = QStyle.SC_None

    @integer
    def initialize(self, low, high):
        """Initialize the Slider with a low and high position and a new range
        """
        assert high >= low, "Maximum range must be larger than minimum range!"

        # Signed 32 bit integer protection as Qt cannot handle. Segfault ...
        if (high - low < 1
                or low <= INT32_MIN // 2
                or high >= INT32_MAX // 2):
            self.setEnabled(False)
            self.setVisible(False)
            return

        self.low_handle_pos = low
        self.high_handle_pos = high
        self.setRange(low, high)

    @integer
    def setMinimum(self, value):
        """Reimplemented method of QSlider"""
        super().setMinimum(value)
        if self.low_handle_pos < value:
            self.low_handle_pos = value
            self.update()

    @integer
    def setMaximum(self, value):
        """Reimplemented method of QSlider"""
        super().setMaximum(value)
        if self.high_handle_pos > value:
            self.high_handle_pos = value
            self.update()

    @integer
    def setValue(self, low, high):
        """Reimplemented method of QSlider

        This method safely accounts for the minimum and maximum.
        """
        # Basic level order validation
        low_position = min(low, high)
        high_position = max(low, high)
        # Protect against Qt Overflow ...
        if (low <= INT32_MIN // 2
                or high >= INT32_MAX // 2):
            return
        if low_position < self.minimum():
            low_position = self.minimum()
        if high_position > self.maximum():
            high_position = self.maximum()

        # Set values and update!
        self.low_handle_pos = low_position
        self.high_handle_pos = high_position
        self.update()

    def value(self):
        """Reimplemented method of QSlider"""
        return self.low_handle_pos, self.high_handle_pos

    setSliderPosition = setValue
    sliderPosition = value

    # -----------------------------------------------------------------------
    # Private Interface

    def pick(self, qpoint):
        """Reimplemented private function of QSlider"""
        if self.orientation() == Qt.Horizontal:
            return qpoint.x()
        else:
            return qpoint.y()

    def pixelPosToRangeValue(self, pos):
        """Reimplemented private function of QSlider"""
        style_option = QStyleOptionSlider()
        self.initStyleOption(style_option)
        style = self.style()
        gr = style.subControlRect(
            style.CC_Slider, style_option, style.SC_SliderGroove, self)
        sr = style.subControlRect(
            style.CC_Slider, style_option, style.SC_SliderHandle, self)

        if self.orientation() == Qt.Horizontal:
            sliderLength = sr.width()
            sliderMin = gr.x()
            sliderMax = gr.right() - sliderLength + 1
        else:
            sliderLength = sr.height()
            sliderMin = gr.y()
            sliderMax = gr.bottom() - sliderLength + 1

        slider_span = sliderMax - sliderMin

        return style.sliderValueFromPosition(
            self.minimum(), self.maximum(), pos - sliderMin, slider_span,
            style_option.upsideDown)

    # -----------------------------------------------------------------------
    # Paint Methods

    def paintEvent(self, event):
        """Reimplemented method of QSlider"""
        with QPainter(self) as painter:
            style = self.style()
            style_option = QStyleOptionSlider()
            self.initStyleOption(style_option)

            # Lower position
            style_option.rect = self.rect()

            # Draw the groove but no ticks
            style_option.subControls = QStyle.SC_SliderGroove
            style.drawComplexControl(QStyle.CC_Slider, style_option,
                                     painter, self)

            style_option.sliderPosition = self.low_handle_pos
            low_rect = style.subControlRect(QStyle.CC_Slider, style_option,
                                            QStyle.SC_SliderHandle)

            style_option.sliderPosition = self.high_handle_pos
            high_rect = style.subControlRect(QStyle.CC_Slider, style_option,
                                             QStyle.SC_SliderHandle)

            low_pos = self.pick(low_rect.center())
            high_pos = self.pick(high_rect.center())

            min_pos = min(low_pos, high_pos)
            max_pos = max(low_pos, high_pos)

            # Draw span rect between handles
            center = QRect(low_rect.center(), high_rect.center()).center()
            if style_option.orientation == Qt.Horizontal:
                span_rect = QRect(QPoint(min_pos, center.y() - 2),
                                  QPoint(max_pos, center.y() + 1))
            else:
                span_rect = QRect(QPoint(center.x() - 2, min_pos),
                                  QPoint(center.x() + 1, max_pos))

            groove = style.subControlRect(QStyle.CC_Slider, style_option,
                                          QStyle.SC_SliderGroove)
            if style_option.orientation == Qt.Horizontal:
                groove.adjust(0, 0, -1, 0)
            else:
                groove.adjust(0, 0, 0, -1)

            color = self.palette().color(QPalette.Highlight)
            if not self.isEnabled():
                color = color.darker(90)
            painter.setBrush(QBrush(color))
            painter.setPen(QPen(color, 0))
            painter.drawRect(span_rect.intersected(groove))

            for index, value in enumerate([self.low_handle_pos,
                                           self.high_handle_pos]):
                handle = SliderHandle(index)
                style_option.subControls = QStyle.SC_SliderHandle
                style_option.sliderPosition = value
                if handle is self.pressedHandle:
                    style_option.state |= QStyle.State_Sunken
                self.style().drawComplexControl(QStyle.CC_Slider, style_option,
                                                painter, self)

    # -----------------------------------------------------------------------
    # Mouse Events

    def mousePressEvent(self, event):
        """Reimplemented method of QSlider"""
        button = event.button()
        if button == Qt.LeftButton:
            event.accept()
            # We are pressing the slider, either handle or bar
            style_option = QStyleOptionSlider()
            self.initStyleOption(style_option)

            # Check if a dedicated handle has been pressed
            style = self.style()
            event_pos = event.pos()
            for index, value in enumerate(
                    [self.low_handle_pos, self.high_handle_pos]):
                style_option.sliderPosition = value
                hit = style.hitTestComplexControl(
                    style.CC_Slider, style_option, event_pos, self)
                if hit == style.SC_SliderHandle:
                    self.pressedControl = QStyle.SC_SliderHandle
                    self.pressedHandle = SliderHandle(index)
                    sr = style.subControlRect(
                        QStyle.CC_Slider, style_option, QStyle.SC_SliderHandle,
                        self)
                    new_pos = self.pick(event_pos - sr.topLeft())
                    self.clickOffset = self.pixelPosToRangeValue(new_pos)
                    break
            else:
                # SCROLL_BAR handles, we move the bar or a single handle!
                self.pressedHandle = SliderHandle.SCROLL_BAR
                self.pressedControl = QStyle.SC_SliderGroove
                self.clickOffset = self.pixelPosToRangeValue(
                    self.pick(event_pos))
        else:
            event.ignore()

    def mouseReleaseEvent(self, event):
        """Reimplemented method of QSlider"""
        if self.pressedControl == QStyle.SC_None:
            event.ignore()
            return

        self.pressedHandle = None
        self.pressedControl = QStyle.SC_None

    def mouseMoveEvent(self, event):
        """Reimplemented method of QSlider"""
        if self.pressedControl not in (QStyle.SC_SliderHandle,
                                       QStyle.SC_SliderGroove):
            event.ignore()
            return

        style_option = QStyleOptionSlider()
        self.initStyleOption(style_option)

        position = self.pixelPosToRangeValue(self.pick(event.pos()))
        # We are moving the full bar
        if self.pressedControl == QStyle.SC_SliderGroove:
            offset = position - self.clickOffset
            self.high_handle_pos += offset
            self.low_handle_pos += offset
            # Note: Check if we exceed a value of the range slider. This can
            # happen when the SCROLL_BAR is moved and the mouse cursor exceeds
            # the slider
            minimum = self.minimum()
            if self.low_handle_pos < minimum:
                difference = minimum - self.low_handle_pos
                self.low_handle_pos += difference
                self.high_handle_pos += difference

            maximum = self.maximum()
            if self.high_handle_pos > maximum:
                difference = maximum - self.high_handle_pos
                self.low_handle_pos += difference
                self.high_handle_pos += difference

        elif self.pressedHandle is SliderHandle.LOW:
            self.low_handle_pos = position
            # Low slider passed the high slider, align!
            if position > self.high_handle_pos:
                self.high_handle_pos = position
        else:
            self.high_handle_pos = position
            # High slider passed the low slider, align!
            if position < self.low_handle_pos:
                self.low_handle_pos = position

        event.accept()
        self.clickOffset = position
        self.update()
        self.sliderMoved.emit(self.low_handle_pos, self.high_handle_pos)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QLabel

from karabo.native import Configurable, String
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..lamp import LampWidget


class MockLabel(QLabel):
    def setPixmap(self, pixmap):
        self.pixmap = pixmap

    def setMaximumWidth(self, value):
        self.width = value

    def setMaximumHeight(self, value):
        self.height = value


class Object(Configurable):
    state = String()


def test_set_values(gui_app, mocker):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "state")
    target = "karabogui.controllers.display.lamp.QLabel"
    mocker.patch(target, new=MockLabel)

    states = ("CHANGING", "ACTIVE", "PASSIVE", "DISABLED", "STATIC",
              "RUNNING", "NORMAL", "ERROR", "INIT", "UNKNOWN")

    controller = LampWidget(proxy=proxy)
    controller.create(None)

    for state in states:
        set_proxy_value(proxy, "state", state)
        assert controller.widget.pixmap is not None
        controller.widget.pixmap = None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QItemSelectionModel, QPoint, Qt
from qtpy.QtGui import QColor

from karabo.common.api import AlarmCondition, State
from karabo.native import (
    AccessLevel, AccessMode, Bool, Configurable, Float, Int32, Slot, String,
    VectorFloat, VectorHash)
from karabogui.binding.api import (
    DeviceProxy, ProjectDeviceProxy, ProxyStatus, apply_default_configuration,
    build_binding)
from karabogui.configurator.api import (
    ConfigurationTreeModel, ConfigurationTreeView, ConfiguratorFilterModel)
from karabogui.indicators import PROPERTY_READONLY_COLOR
from karabogui.testing import GuiTestCase


def is_editable_index(index):
    return index.flags() & Qt.ItemIsEditable == Qt.ItemIsEditable


class RowSchema(Configurable):
    start = Float()
    stop = Float()


class Object(Configurable):
    state = String(
        displayedName="State",
        enum=State,
        displayType="State",
        defaultValue=State.UNKNOWN,
        accessMode=AccessMode.READONLY)

    foo = Bool(
        displayedName="Foo",
        defaultValue=True)

    bar = Float(
        defaultValue=1.2,
        minInc=0.0,
        maxInc=10.0)

    baz = Int32()

    qux = String(
        options=["foo", "bar", "baz", "qux"])

    vector = VectorFloat(
        defaultValue=[1.2],
        minSize=1, maxSize=2)

    table = VectorHash(
        displayedName="Table",
        rows=RowSchema)

    alarmCondition = String(
        defaultValue=AlarmCondition.NONE,
        displayedName="Alarm",
        enum=AlarmCondition,
        displayType="AlarmCondition",
        accessMode=AccessMode.READONLY)

    expertProp = Float(
        defaultValue=1.2,
        requiredAccessLevel=AccessLevel.EXPERT)

    @Slot(allowedStates=[State.INTERLOCKED, State.ACTIVE])
    def setSpeed(self):
        """Simple slot dummy method"""


class TestConfiguratorProjectDevice(GuiTestCase):
    def setUp(self):
        super().setUp()
        self.view = ConfigurationTreeView()
        binding = build_binding(Object.getClassSchema())
        root = ProjectDeviceProxy(binding=binding, server_id="Test",
                                  status=ProxyStatus.OFFLINE)
        apply_default_configuration(root.binding)
        self.view.assign_proxy(proxy=root)
        self.model = self.view.model()
        assert isinstance(self.model, ConfigurationTreeModel)
        self.model.root = root
        assert self.model._is_root_class
        self.model._config_update()

    def test_basics(self):
        # This is constant
        assert self.model.columnCount() == 3
        # The number of properties in the Object class up above
        assert self.model.rowCount() == 9

    def test_get_property_proxy_data(self):
        state_index = self.model.index(0, 0)
        assert state_index.data() == "State"
        assert state_index.data(role=Qt.ToolTipRole) is None

        state_index_value = self.model.index(0, 1)
        assert state_index_value.data(Qt.EditRole) is None

        bar_index = self.model.index(2, 0)
        assert bar_index.data() == "bar"
        text = "Key: bar - AccessLevel: OPERATOR - Access Allowed: True"
        assert bar_index.data(role=Qt.ToolTipRole) == text
        bar_index = self.model.index(2, 2)
        text = "Key: bar - AccessLevel: OPERATOR - Access Allowed: True"
        assert bar_index.data(role=Qt.ToolTipRole) == text

        proxy = self.model.property_proxy("bar")
        assert proxy.binding is not None
        assert proxy.value == 1.2
        bar_index = self.model.index(2, 1)
        assert bar_index.data() == "1.2"

        self.model.setData(bar_index, 3.2, role=Qt.EditRole)
        assert bar_index.data() == "3.2"
        self.model.setData(bar_index, 1.2, role=Qt.EditRole)
        assert bar_index.data() == "1.2"

        # Test via flushing, we have a project device
        proxy.edit_value = 5.2
        self.model.flush_index_modification(bar_index)
        assert bar_index.data() == "5.2"

        proxy.edit_value = 7.2
        self.model.clear_index_modification(bar_index)
        assert proxy.edit_value is None
        assert bar_index.data() == "5.2"

        assert bar_index.data(role=Qt.BackgroundRole) is None

    def test_attributes(self):
        bar_index = self.model.index(2, 0)
        assert bar_index.isValid()
        assert bar_index.data() == "bar"

        # no more attributes
        assert self.model.rowCount(parent=bar_index) == 0

        # ------------------------------------
        # Test the vector
        vector_index = self.model.index(5, 0)
        assert vector_index.isValid()
        assert vector_index.data() == "vector"
        # No attribute left
        assert self.model.rowCount(parent=vector_index) == 0

    def test_flags(self):
        bar_index = self.model.index(2, 0)
        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled
        assert bar_index.flags() == flags

    def test_parent(self):
        bar_index = self.model.index(2, 0)
        assert not bar_index.parent().isValid()

    def test_swap(self):
        self.view.swap_models()
        assert isinstance(self.view.model(), ConfiguratorFilterModel)
        assert not self.view.model().dynamicSortFilter()
        self.view.swap_models()
        assert not isinstance(self.view.model(), ConfiguratorFilterModel)


class TestConfiguratorDevice(GuiTestCase):
    def setUp(self):
        super().setUp()
        self.view = ConfigurationTreeView()
        binding = build_binding(Object.getClassSchema())
        root = DeviceProxy(binding=binding, server_id="Test",
                           status=ProxyStatus.ONLINE)
        apply_default_configuration(root.binding)
        self.view.assign_proxy(proxy=root)
        self.model = self.view.model()
        assert isinstance(self.model, ConfigurationTreeModel)
        self.model.root = root

    def test_get_property_proxy_data(self):
        state_index = self.model.index(0, 0)
        assert state_index.data() == "State"
        proxy = self.model.property_proxy("state")
        assert proxy.binding is not None
        assert proxy.value == "UNKNOWN"
        state_index = self.model.index(0, 1)
        assert state_index.data() == "UNKNOWN"
        brush = state_index.data(role=Qt.BackgroundRole)
        assert brush is not None
        color = brush.color()
        assert color.blue() == 0
        assert color.red() == 255
        assert color.green() == 170

        # Device updates from outside
        proxy.binding.trait_setq(value="ON")
        brush = state_index.data(role=Qt.BackgroundRole)
        assert state_index.data() == "ON"
        assert brush is not None
        color = brush.color()
        assert color.blue() == 0
        assert color.red() == 120
        assert color.green() == 255

        # Alarm coloring
        alarm_index = self.model.index(7, 0)
        assert alarm_index.data() == "Alarm"
        proxy = self.model.property_proxy("alarmCondition")
        assert proxy.binding is not None
        assert proxy.value == "none"
        alarm_index = self.model.index(7, 1)
        assert alarm_index.data() == "none"
        brush = alarm_index.data(role=Qt.BackgroundRole)
        assert brush is not None
        color = brush.color()
        assert color.blue() == 225
        assert color.red() == 225
        assert color.green() == 242

        # Test apply all and decline all, we have an online device
        proxy = self.model.property_proxy("bar")
        assert proxy.binding is not None
        assert proxy.value == 1.2
        bar_index = self.model.index(2, 1)
        assert bar_index.data() == "1.2"
        proxy.edit_value = 5.2
        path = "karabogui.configurator.qt_item_model.send_property_changes"
        with mock.patch(path) as changes:
            self.view.apply_all()
            changes.assert_called_once_with([proxy])

        self.view.decline_all()
        assert proxy.edit_value is None

        proxy.edit_value = 3.4
        self.view._reset_to_default(proxy)
        assert proxy.edit_value == 1.2

        # Test the context menu, we have an online device, but take an edit
        # index!
        edit_index = self.model.index(2, 2)
        self.view.selectionModel().setCurrentIndex(
            edit_index, QItemSelectionModel.ClearAndSelect)
        with mock.patch("karabogui.configurator.view.QMenu") as menu:
            self.view._show_context_menu(None)
            menu.assert_called_once()

        # Show the pop up widget
        event_pos = QPoint(20, 20)
        with mock.patch("karabogui.configurator.view.PopupWidget") as widget:
            self.view._show_popup_widget(self.model.index(0, 0), event_pos)
            widget.assert_called_once()
            widget().setInfo.assert_called_with(
                dict([('Property', 'State'), ('Key', 'state'),
                      ('Value Type', 'String'),
                      ('Default Value', 'UNKNOWN'),
                      ('AccessMode', 'READONLY'),
                      ('AccessLevel', 'OBSERVER'),
                      ('Assignment', 'OPTIONAL'),
                      ('Value on device', 'ON'),
                      ('displayType', "State"),
                      ('metricPrefixSymbol', ''),
                      ('unitSymbol', ''),
                      ]))
            assert self.view.popup_widget is not None

    def test_modeltester_qt(self):
        from pytestqt.modeltest import ModelTester

        binding = build_binding(Object.getClassSchema())
        for proxy_type in (DeviceProxy, ProjectDeviceProxy):
            root = proxy_type(binding=binding, server_id="Test",
                              status=ProxyStatus.ONLINE)
            apply_default_configuration(root.binding)
            self.view.assign_proxy(proxy=root)
            self.model.root = root
            tester = ModelTester(None)
            tester.check(self.model)


class TestConfiguratorProjectDeviceExpert(GuiTestCase):
    """Check the offline configuration scenario"""

    def setUp(self):
        super().setUp()
        self.view = ConfigurationTreeView()
        binding = build_binding(Object.getClassSchema())
        root = ProjectDeviceProxy(binding=binding, server_id="Test",
                                  status=ProxyStatus.OFFLINE)
        apply_default_configuration(root.binding)
        self.view.assign_proxy(proxy=root)
        self.model = self.view.model()
        assert isinstance(self.model, ConfigurationTreeModel)
        self.model.root = root
        assert self.model._is_root_class
        self.model._config_update()
        # Set the EXPERT MODE
        self.view.setMode(True)

    def test_basics(self):
        # This is constant
        assert self.model.columnCount() == 3
        # One more due to EXPERT mode
        assert self.model.rowCount() == 10

    def test_get_property_proxy_data(self):
        state_index = self.model.index(0, 0)
        assert state_index.data() == "State"
        assert state_index.data(role=Qt.ToolTipRole) is None

        state_index_value = self.model.index(0, 1)
        assert state_index_value.data(Qt.EditRole) is None

        # Editable
        bar_index = self.model.index(2, 0)
        assert bar_index.data() == "bar"
        bar_index_value = self.model.index(2, 2)
        assert is_editable_index(bar_index_value)
        color = bar_index_value.data(Qt.ForegroundRole)
        assert color is None

        expert_index = self.model.index(8, 0)
        assert expert_index.data() == "expertProp"
        assert expert_index.data(Qt.EditRole) is None
        # Edit column
        expert_index_value = self.model.index(8, 2)
        assert not is_editable_index(expert_index_value)
        assert expert_index_value.data(Qt.DisplayRole) == "1.2"
        assert expert_index_value.data(Qt.EditRole) == "1.2"
        color = expert_index_value.data(Qt.ForegroundRole)
        assert color == QColor(*PROPERTY_READONLY_COLOR)


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from functools import partial

import karabogui as gui_pkg
from karabo.testing.import_checker import (
    check_for_disallowed_module_imports, check_for_star_imports,
    run_checker_on_package)


def test_no_star_imports():
    run_checker_on_package(gui_pkg, check_for_star_imports)


def test_no_middlelayer_imports():
    for forbidden in ("karabo.middlelayer", "karabo.middlelayer_api"):
        checker = partial(check_for_disallowed_module_imports, forbidden)
        run_checker_on_package(gui_pkg, checker)


def test_no_gui_binding_imports():
    for forbidden in ("PyQt5", "PyQt6", "PySide", "PySide2"):
        checker = partial(check_for_disallowed_module_imports, forbidden)
        run_checker_on_package(gui_pkg, checker)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Slot

from karabogui.graph.common.api import (
    AuxPlots, ExportTool, ExportToolset, ROITool, ROIToolset,
    ToolbarController)
from karabogui.graph.image.tools.toolbar import AuxPlotsToolset
from karabogui.testing import GuiTestCase


class TestCase(GuiTestCase):
    def setUp(self):
        super().setUp()
        self.toolbar = ToolbarController()

    def test_add_aux_plots_toolset(self):
        self.toolbar.register_toolset(toolset=AuxPlots, klass=AuxPlotsToolset)
        toolset = self.toolbar.add_toolset(AuxPlots)
        assert len(self.toolbar.toolsets) == 2
        assert AuxPlots in self.toolbar.toolsets
        assert isinstance(self.toolbar.toolsets[AuxPlots],
                          AuxPlotsToolset)
        assert toolset is self.toolbar.toolsets[AuxPlots]

    def test_add_roi_toolset(self):
        toolset = self.toolbar.add_toolset(ROITool)
        assert len(self.toolbar.toolsets) == 2
        assert ROITool in self.toolbar.toolsets
        assert isinstance(self.toolbar.toolsets[ROITool], ROIToolset)
        assert toolset is self.toolbar.toolsets[ROITool]

    def test_add_export_toolset(self):
        toolset = self.toolbar.add_toolset(ExportTool)
        assert len(self.toolbar.toolsets) == 2
        assert ExportTool in self.toolbar.toolsets
        assert isinstance(self.toolbar.toolsets[ExportTool],
                          ExportToolset)
        assert toolset is self.toolbar.toolsets[ExportTool]


class TestAuxPlotsToolset(GuiTestCase):
    def setUp(self):
        super().setUp()
        self.toolset = AuxPlotsToolset()

    def test_button_click(self):
        # Set up test
        tool = None

        @Slot(object)
        def set_tool(value):
            nonlocal tool
            tool = value

        self.toolset.on_trait_change(set_tool, "current_tool")
        profile_plot_button = self.toolset.buttons[AuxPlots.ProfilePlot]

        # Check if profile plot is checked by default
        assert not profile_plot_button.isChecked()

        # Click profile plot button for the first time
        profile_plot_button.click()
        assert tool is AuxPlots.ProfilePlot
        assert profile_plot_button.isChecked()

        # Click profile plot button again
        profile_plot_button.click()
        assert tool is AuxPlots.NoPlot
        assert not profile_plot_button.isChecked()

    def test_check(self):
        profile_plot_button = self.toolset.buttons[AuxPlots.ProfilePlot]

        # Check if profile plot is checked by default
        assert not profile_plot_button.isChecked()

        # Set the button as checked
        self.toolset.check(AuxPlots.ProfilePlot)
        assert profile_plot_button.isChecked()

    def test_uncheck_all(self):
        for button in self.toolset.buttons.values():
            assert not button.isChecked()

        # Set one button as checked just to test
        button.setChecked(True)

        # Uncheck all buttons
        self.toolset.uncheck_all()
        for button in self.toolset.buttons.values():
            assert not button.isChecked()


class TestROIToolset(GuiTestCase):
    def setUp(self):
        super().setUp()
        self.toolset = ROIToolset()

    def test_button_click(self):
        # Set up test
        tool = None

        @Slot(object)
        def set_tool(value):
            nonlocal tool
            tool = value

        self.toolset.on_trait_change(set_tool, "current_tool")

        # Check if ROI buttons are unchecked by default
        for button in self.toolset.buttons.values():
            assert not button.isChecked()

        # Click Rect ROI for the first time
        self.toolset.buttons[ROITool.Rect].click()
        assert tool is ROITool.Rect
        assert self.toolset.buttons[ROITool.Rect].isChecked()
        assert not self.toolset.buttons[ROITool.Crosshair].isChecked()

        # Click Crosshair ROI for the first time
        self.toolset.buttons[ROITool.Crosshair].click()
        assert tool is ROITool.Crosshair
        assert self.toolset.buttons[ROITool.Crosshair].isChecked()
        assert not self.toolset.buttons[ROITool.Rect].isChecked()

        # Click Crosshair ROI again
        self.toolset.buttons[ROITool.Crosshair].click()
        assert tool is ROITool.NoROI
        assert not self.toolset.buttons[ROITool.Crosshair].isChecked()
        assert not self.toolset.buttons[ROITool.Rect].isChecked()

    def test_check_button(self):
        rect_roi_button = self.toolset.buttons[ROITool.Rect]

        # Check if profile plot is checked by default
        assert not rect_roi_button.isChecked()

        # Set the button as checked
        self.toolset.check(ROITool.Rect)
        assert rect_roi_button.isChecked()

    def test_uncheck_all(self):
        for button in self.toolset.buttons.values():
            assert not button.isChecked()

        # Set one button as checked just to test
        button.setChecked(True)

        # Uncheck all buttons
        self.toolset.uncheck_all()
        for button in self.toolset.buttons.values():
            assert not button.isChecked()



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on October 25, 2022
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy import uic
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog

from karabo.native import create_html_hash, dictToHash

from .utils import get_dialog_ui


class DataViewDialog(QDialog):
    def __init__(self, title, info, data, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(False)
        uic.loadUi(get_dialog_ui("data_preview_dialog.ui"), self)
        self.setWindowTitle(title)
        flags = Qt.WindowCloseButtonHint | Qt.WindowStaysOnTopHint
        self.setWindowFlags(self.windowFlags() | flags)

        if info is None:
            self.ui_info.setVisible(False)
        else:
            self.ui_info.setText(info)
        data = dictToHash(data)
        html = create_html_hash(data, include_attributes=False)
        self.ui_text_info.setHtml(html)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from numpy.testing import assert_allclose
from qtpy.QtCore import Slot

from karabogui.graph.image.dialogs.levels import LevelsDialog, get_decimals
from karabogui.graph.image.plot import KaraboImagePlot
from karabogui.testing import GuiTestCase


class TestLevelsDialog(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.plotItem = KaraboImagePlot()
        self.imageItem = self.plotItem.imageItem

    def tearDown(self):
        super().tearDown()
        self.imageItem.deleteLater()
        self.imageItem = None

    def test_image_preview_signal(self):
        image = np.random.randint(-100, 100, (100, 100))
        image_range = image.min(), image.max()

        self.imageItem.setImage(image)
        image_levels = self.imageItem.levels
        auto_levels = self.imageItem.auto_levels
        assert auto_levels
        dialog = LevelsDialog(image_levels, image_range, auto_levels)

        # ----------------------------------------------------------------
        # Change auto level and cross check with a slot

        assert dialog.automatic_checkbox.isChecked()
        assert not dialog.min_spinbox.isEnabled()
        assert not dialog.max_spinbox.isEnabled()

        received = None
        count = 0

        @Slot()
        def levelUpdate(levels):
            nonlocal received, count
            received = levels
            count += 1

        dialog.levelsPreview.connect(levelUpdate)

        # Now change the auto levels
        assert received is None, None
        dialog.automatic_checkbox.setChecked(False)
        assert dialog.levels is not None, None
        assert count == 1
        assert received is not None
        dialog.automatic_checkbox.setChecked(True)
        assert dialog.levels is None
        assert count == 2
        assert received is None

        # ----------------------------------------------------------------
        # Change spin boxes

        dialog.automatic_checkbox.setChecked(False)
        assert dialog.min_spinbox.isEnabled()
        assert dialog.max_spinbox.isEnabled()
        count = 0

        dialog.min_spinbox.setValue(2)
        assert count == 1
        assert received == [2.0, 99.0]
        dialog.max_spinbox.setValue(55)
        assert count == 2
        assert received == [2.0, 55.0]

        # For now, `slider.setValue` does not emit
        dialog.slider.setValue(13, 44)
        assert count == 2

        # Finally disconnect
        dialog.levelsPreview.disconnect(levelUpdate)

    def test_image_autolevel(self):
        """Tests the dialog initialization for imageItems with autolevels"""

        # Setup levels dialog
        image = np.random.randint(-100, 100, (100, 100))
        image_range = image.min(), image.max()

        self.imageItem.setImage(image)
        image_levels = self.imageItem.levels
        auto_levels = self.imageItem.auto_levels

        # Do some prior checking before the actual tests
        np.testing.assert_array_equal(image_levels, image_range)
        assert auto_levels

        # Instantiate dialog
        dialog = LevelsDialog(image_levels, image_range, auto_levels)

        # Check dialog widgets if enabled
        assert dialog.automatic_checkbox.isChecked()
        assert not dialog.min_spinbox.isEnabled()
        assert not dialog.max_spinbox.isEnabled()

        # Check dialog widget values
        assert dialog.min_spinbox.value() == image_levels[0]
        assert dialog.max_spinbox.value() == image_levels[1]

        # Check slider range
        assert not dialog.slider.isEnabled()
        assert dialog.slider.minimum() == image_range[0]
        assert dialog.slider.maximum() == image_range[1]

    def test_image_float_range(self):
        # Setup levels dialog
        image = np.random.randn(100, 100)
        image_range = image.min(), image.max()

        self.imageItem.setImage(image)
        image_levels = self.imageItem.levels
        auto_levels = self.imageItem.auto_levels

        # Do some prior checking before the actual tests
        np.testing.assert_array_equal(image_levels, image_range)
        assert auto_levels

        # Instantiate dialog
        dialog = LevelsDialog(image_levels, image_range, auto_levels)

        # Check dialog widgets if enabled
        assert dialog.automatic_checkbox.isChecked()
        assert not dialog.min_spinbox.isEnabled()
        assert not dialog.max_spinbox.isEnabled()

        # Check dialog widget values
        assert_allclose(dialog.min_spinbox.value(), image_levels[0], atol=0.1)
        assert_allclose(dialog.max_spinbox.value(), image_levels[1], atol=0.1)

        # Check slider range
        assert not dialog.slider.isEnabled()

    def test_image_not_autolevel(self):
        """Tests the dialog initialization for imageItems with
           enforced levels"""

        # Setup levels dialog
        image = np.random.randint(-100, 100, (100, 100))
        input_levels = [-50, 50]

        self.imageItem.setImage(image)
        self.plotItem.set_image_levels(input_levels)
        image_levels = self.imageItem.levels
        auto_levels = self.imageItem.auto_levels

        # Do some prior checking before the actual tests
        np.testing.assert_array_equal(image_levels, input_levels)
        assert not auto_levels

        # Instantiate dialog
        dialog = LevelsDialog(image_levels, input_levels, auto_levels)

        # Check dialog widgets if enabled
        assert not dialog.automatic_checkbox.isChecked()
        assert dialog.min_spinbox.isEnabled()
        assert dialog.max_spinbox.isEnabled()

        # Check dialog widget values
        assert dialog.min_spinbox.value() == image_levels[0]
        assert dialog.min_label.text() == str(image_levels[0])
        assert dialog.max_spinbox.value() == image_levels[1]
        assert dialog.max_label.text() == str(image_levels[1])

    def test_accepted_autolevel(self):
        """Check if accepted values are input values"""

        # Setup levels dialog
        image = np.random.randint(-100, 100, (100, 100))
        image_range = image.min(), image.max()

        input_levels = [-50, 50]
        self.imageItem.setImage(image)
        self.plotItem.set_image_levels(input_levels)

        # Instantiate dialog
        dialog = LevelsDialog(self.imageItem.levels, image_range,
                              self.imageItem.auto_levels)

        # Change the values then accept the dialog
        dialog.automatic_checkbox.setChecked(True)
        dialog.accept()

        # Check if changes are saved
        assert dialog.levels is None

    def test_accepted_not_autolevel(self):
        # Setup levels dialog
        image = np.random.randint(-100, 100, (100, 100))
        image_range = image.min(), image.max()

        self.imageItem.setImage(image)

        # Instantiate dialog
        dialog = LevelsDialog(self.imageItem.levels, image_range,
                              self.imageItem.auto_levels)

        # Change the values then accept the dialog
        input_levels = [-50, 50]
        dialog.automatic_checkbox.setChecked(False)
        dialog.min_spinbox.setValue(input_levels[0])
        dialog.max_spinbox.setValue(input_levels[1])
        dialog.accept()

        # Check if changes are saved
        np.testing.assert_array_equal(dialog.levels, input_levels)

    def test_accepted_not_autolevel_false_input(self):
        # Setup levels dialog
        image = np.random.randint(-100, 100, (100, 100))
        image_range = image.min(), image.max()

        self.imageItem.setImage(image)
        min_limit = -70
        max_limit = 200

        preview_levels = None

        @Slot(object)
        def preview(levels):
            nonlocal preview_levels
            preview_levels = levels

        # Instantiate dialog
        dialog = LevelsDialog(self.imageItem.levels, image_range,
                              self.imageItem.auto_levels,
                              limits=[min_limit, max_limit])
        dialog.levelsPreview.connect(preview)

        # Check the min and max protection
        assert dialog.min_spinbox.minimum() == min_limit
        assert dialog.min_spinbox.maximum() == max_limit
        assert dialog.max_spinbox.minimum() == min_limit
        assert dialog.max_spinbox.maximum() == max_limit

        # Change the values then accept the dialog
        input_levels = [-50, 50]
        dialog.automatic_checkbox.setChecked(False)
        dialog.min_spinbox.setValue(input_levels[1])
        dialog.max_spinbox.setValue(input_levels[0])
        assert preview_levels == [-50, 50]
        dialog.accept()

        # Check if changes are saved
        np.testing.assert_array_equal(dialog.levels, input_levels)


def test_get_decimals():
    assert get_decimals([-10.0, 1.0]) == 1
    assert get_decimals([-10.0, 10.0]) == 1
    assert get_decimals([1e-4, 1e-5]) == 7
    assert get_decimals([1e4, 1e-5]) == 1
    assert get_decimals([0.021, 0.01]) == 4
    assert get_decimals([0.0, 0.01]) == 1
    assert get_decimals([1e4, 1e5]) == 1
    assert get_decimals([0, 10.0]) == 1
    assert get_decimals([-10.0, 0.0]) == 1



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Qt, Slot
from qtpy.QtGui import QPalette
from qtpy.QtWidgets import QLineEdit


class LineEditEditor(QLineEdit):
    """A colored `LineEdit` depending on acceptable input

    Note: This widget is included as promoted widgets in *.ui files.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._normal_palette = self.palette()
        self._error_palette = QPalette(self._normal_palette)
        self._error_palette.setColor(QPalette.Text, Qt.red)
        self.textChanged.connect(self._check_background)

    @Slot(str)
    def _check_background(self, text):
        acceptable_input = self.hasAcceptableInput()
        palette = (self._normal_palette if acceptable_input
                   else self._error_palette)
        self.setPalette(palette)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on April 15, 2021
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from copy import deepcopy

from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QDialog, QDialogButtonBox

from karabo.native import Hash, create_html_hash
from karabogui.binding.api import (
    ProxyStatus, extract_init_configuration, validate_binding_configuration)

from .utils import get_dialog_ui

_NO_SCHEMA_STATUS = (ProxyStatus.MISSING, ProxyStatus.NOSERVER,
                     ProxyStatus.NOPLUGIN)


class DeviceConfigurationDialog(QDialog):
    """The basic device configuration dialog for a project device

    This QDialog shows the `Hash` configuration in a html format and can
    show the conflicts on request.
    """

    def __init__(self, name, configuration, project_device=None,
                 editable=False, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(False)
        uic.loadUi(get_dialog_ui("device_configuration.ui"), self)
        self.setWindowTitle(f"Configuration {name} "
                            f"- {project_device.device_id}")
        flags = Qt.WindowCloseButtonHint | Qt.WindowStaysOnTopHint
        self.setWindowFlags(self.windowFlags() | flags)

        apply_button = self.ui_buttonBox.button(QDialogButtonBox.Apply)
        apply_button.clicked.connect(self.accept)
        discard_button = self.ui_buttonBox.button(QDialogButtonBox.Close)
        discard_button.clicked.connect(self.reject)
        discard_button.setFocus()

        if not editable:
            self.ui_edit_group.setVisible(False)
            self.ui_buttonBox.button(QDialogButtonBox.Apply).setVisible(False)

        self._invalid_paths = []
        self.ui_erase_path.clicked.connect(self._erase_path)
        self.ui_erase_all.clicked.connect(self._erase_config)
        self.ui_cleanup.clicked.connect(self._cleanup_configuration)
        if project_device.status in _NO_SCHEMA_STATUS:
            self.ui_cleanup.setEnabled(False)
            text = "Cleanup is only available with a class schema"
            self.ui_cleanup.setToolTip(text)
        self.project_device = project_device
        # Make sure we deepcopy for manipulations!
        self.configuration = deepcopy(configuration)
        self._show_configuration()

    @Slot()
    def _cleanup_configuration(self):
        config = Hash()
        for k, v, a in Hash.flat_iterall(self.configuration, empty=False):
            if k in self._invalid_paths:
                continue
            config[k] = v

        proxy = self.project_device.get_class_proxy()
        binding = proxy.binding
        # Schema is available, filter out read only properties without
        # run time attributes and remove attrs that are not runtime attributes
        if len(binding.value):
            config = extract_init_configuration(
                binding=binding, config=config)

        self.configuration = config
        self._show_configuration()

    @Slot()
    def _erase_config(self):
        """Clear the full configuration"""
        self.configuration = Hash()
        self._show_configuration()

    @Slot()
    def _erase_path(self):
        path = self.ui_paths.currentText()
        index = self.ui_paths.currentIndex()
        if path in self.configuration:
            self.configuration.erase(path)
        self._show_configuration()
        if self.ui_paths.count() > index:
            self.ui_paths.setCurrentIndex(index)

    @Slot()
    def _show_configuration(self):
        self._update_configuration_view()
        self._extract_invalid_paths()

    def _update_configuration_view(self):
        widget = self.ui_text_info
        config = self.configuration
        # Note: Protect as well against a `None` configuration!
        if config is None or config.empty():
            html = "<center>No changes in configuration!</center>"
            self.ui_paths.clear()
            self.ui_paths.setEnabled(False)
            self.ui_erase_path.setEnabled(False)
            self.ui_num_paths.setText("# 0")
        else:
            html = create_html_hash(config)
            # Enable to erase all path, broken and valid ones!
            self.ui_paths.clear()
            self.ui_paths.setEnabled(True)
            self.ui_erase_path.setEnabled(True)
            paths = config.paths(intermediate=True)
            self.ui_num_paths.setText(f"# {len(paths)}")
            self.ui_paths.addItems(paths)

        widget.setHtml(html)
        scroll_bar = widget.verticalScrollBar()
        scroll_bar.setValue(scroll_bar.sliderPosition())

    def _extract_invalid_paths(self):
        proxy = self.project_device.get_class_proxy()
        binding = proxy.binding
        # No schema arrived yet!
        if not len(binding.value):
            invalid = []
        else:
            config = validate_binding_configuration(
                binding=binding, config=self.configuration)
            if config.empty():
                invalid = []
            else:
                invalid = config.paths(intermediate=False)

        self._invalid_paths = invalid



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Qt
from qtpy.QtGui import QValidator
from qtpy.QtWidgets import QLineEdit
from traits.api import Instance, Int, String, Undefined

from karabogui.binding.api import get_editor_value
from karabogui.controllers.base import BaseBindingController
from karabogui.controllers.unitlabel import add_unit_label
from karabogui.controllers.util import is_proxy_allowed
from karabogui.util import SignalBlocker, generateObjectName
from karabogui.widgets.hints import LineEdit

FINE_COLOR = "black"
ERROR_COLOR = "red"


class BaseLineEditController(BaseBindingController):
    """The `BaseLineEditController` class provides a visual indication
    if a set text provides acceptable input validated by a subclassed
     `validator` instance.
    """
    model = Undefined
    # The line edit widget instance
    internal_widget = Instance(QLineEdit)
    # The validator instance
    validator = Instance(QValidator, allow_none=True)
    # The last cursor position
    last_cursor_pos = Int(0)

    # Private properties
    _style_sheet = String("")

    def create_widget(self, parent):
        self.internal_widget = LineEdit(parent)
        self.validator = self.create_validator()
        self.internal_widget.setValidator(self.validator)
        objectName = generateObjectName(self)
        self._style_sheet = (f"QWidget#{objectName}" +
                             " {{ color: {}; }}")
        self.internal_widget.setObjectName(objectName)
        sheet = self._style_sheet.format(FINE_COLOR)
        self.internal_widget.setStyleSheet(sheet)
        widget = add_unit_label(self.proxy, self.internal_widget,
                                parent=parent)
        widget.setFocusProxy(self.internal_widget)
        return widget

    def set_read_only(self, ro):
        self.internal_widget.setReadOnly(ro)
        if not ro:
            self.internal_widget.textChanged.connect(self._on_text_changed)

        focus_policy = Qt.NoFocus if ro else Qt.StrongFocus
        self.internal_widget.setFocusPolicy(focus_policy)

    def state_update(self, proxy):
        enable = is_proxy_allowed(proxy)
        self.internal_widget.setEnabled(enable)

    def setEnabled(self, enable):
        self.internal_widget.setEnabled(enable)

    def binding_update(self, proxy):
        self.binding_validator(proxy)
        self.widget.update_unit_label(proxy)

    def value_update(self, proxy):
        value = get_editor_value(proxy, "")
        if value != "":
            try:
                value = self.toString(value)
            except Exception:
                value = str(value)
        with SignalBlocker(self.internal_widget):
            self.internal_widget.setText(value)
        self.internal_widget.setCursorPosition(self.last_cursor_pos)

    def on_decline(self):
        """When the input was declined, this method is executed"""
        sheet = self._style_sheet.format(FINE_COLOR)
        self.internal_widget.setStyleSheet(sheet)

    # Private interface
    # ---------------------------------------------------------------------

    def _on_text_changed(self, text):
        if self.proxy.binding is None:
            sheet = self._style_sheet.format(FINE_COLOR)
            self.internal_widget.setStyleSheet(sheet)
            return

        self.onText(text)

    # Public interface
    # ---------------------------------------------------------------------

    def validate_text_color(self):
        """Public method to validate the text color"""
        acceptable_input = self.internal_widget.hasAcceptableInput()
        color = FINE_COLOR if acceptable_input else ERROR_COLOR
        sheet = self._style_sheet.format(color)
        self.internal_widget.setStyleSheet(sheet)

    # Abstract interface
    # ----------------------------------------------------------------------

    def onText(self, text):
        """Subclass method to react on the update of the internal widget"""
        color = FINE_COLOR
        acceptable_input = self.internal_widget.hasAcceptableInput()
        if acceptable_input:
            self.last_cursor_pos = self.internal_widget.cursorPosition()
            self.proxy.edit_value = self.fromString(text)
        else:
            color = ERROR_COLOR
            # Erase the edit value by setting it to `None`
            self.proxy.edit_value = None
        sheet = self._style_sheet.format(color)
        self.internal_widget.setStyleSheet(sheet)

    def create_validator(self):
        """Subclass method to specify a validator instance of `QValidator`"""

    def binding_validator(self, proxy):
        """Subclass method to update a validator on binding update"""

    def toString(self, value):
        """Subclass method to convert a value to string"""
        return str(value)

    def fromString(self, value):
        """Subclass method to convert a value from a string"""
        return value



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabo.common.scenemodel.api import build_model_config
from karabo.native import Bool, Configurable, Int32
from karabogui.binding.api import PropertyProxy, build_binding
from karabogui.testing import (
    get_class_property_proxy, get_property_proxy, set_proxy_value)

from ..multicurvegraph import DisplayMultiCurveGraph


class Object(Configurable):
    reset = Bool()
    index = Int32()
    value = Int32()


@pytest.fixture
def multicurvegraph_setup(gui_app):
    schema = Object.getClassSchema()
    reset = get_class_property_proxy(schema, "reset")
    device = reset.root_proxy

    index = PropertyProxy(root_proxy=device, path="index")
    value = PropertyProxy(root_proxy=device, path="value")

    controller = DisplayMultiCurveGraph(proxy=index)
    controller.create(None)
    controller.visualize_additional_property(reset)
    controller.visualize_additional_property(value)

    yield controller, reset, index, value

    controller.destroy()
    assert controller.widget is None


def test_set_value(multicurvegraph_setup):
    controller, reset, index, value = multicurvegraph_setup
    assert len(controller._curves) == 1
    set_proxy_value(index, "index", 1)
    set_proxy_value(value, "value", 42)
    empty = controller._curves.get(index)
    assert empty is None
    curve = controller._curves.get(value)
    assert curve is not None
    assert list(curve.xData) == [1.0]
    assert list(curve.yData) == [42.0]
    set_proxy_value(index, "index", 2)
    set_proxy_value(value, "value", 37)
    assert list(curve.xData) == [1.0, 2.0]
    assert list(curve.yData) == [42.0, 37.0]

    set_proxy_value(reset, "reset", True)
    curve = controller._curves.get(value)
    assert curve is not None

    empty_x = curve.xData is None or list(curve.xData) == []
    empty_y = curve.yData is None or list(curve.yData) == []
    assert empty_x
    assert empty_y

    set_proxy_value(value, "value", 42)
    curve = controller._curves.get(value)

    empty = curve.yData is None or list(curve.yData) == []
    assert empty
    set_proxy_value(index, "index", 1)
    assert list(curve.yData) == [42.0]


def test_visualize_additional_nobinding(multicurvegraph_setup):
    controller, _, _, value = multicurvegraph_setup
    proxy = get_property_proxy(None, "value", "TestDevice2")
    assert controller.visualize_additional_property(proxy)
    assert len(controller._curves) == 1
    build_binding(Object.getClassSchema(),
                  existing=proxy.root_proxy.binding)
    assert len(controller._curves) == 2


def test_configuration(multicurvegraph_setup):
    controller, _, _, _ = multicurvegraph_setup
    default_config = build_model_config(controller.model)
    configuration = controller.widget.configuration
    assert configuration
    assert configuration == default_config


def test_remove_proxy(multicurvegraph_setup):
    controller, reset, _, value = multicurvegraph_setup
    assert controller.remove_additional_property(reset)
    assert controller._reset_proxy is None

    legend = controller.widget._legend
    assert legend.isVisible()

    assert controller._curves.get(value)
    assert controller.remove_additional_property(value)
    assert controller._curves.get(value) is None
    assert not legend.isVisible()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os
import shutil
import subprocess
import sys
from pathlib import Path
from platform import system

from qtpy.QtCore import QSettings

import karabogui.icons as icons
from karabogui.const import HIDDEN_KARABO_FOLDER
from karabogui.programs.utils import create_linux_desktop_file

ON_WINDOWS = system() == "Windows"
PYTHON_SHEBANG = f"#!{sys.executable}"

PROTOCOL_HANDLER = "protocol_handler.py"
PROTOCOL_HANDLER_SRC = str(Path(__file__).parent / PROTOCOL_HANDLER)
_dest = Path(HIDDEN_KARABO_FOLDER) / PROTOCOL_HANDLER
if ON_WINDOWS:
    _dest = _dest.with_suffix(".pyw")
PROTOCOL_HANDLER_DEST = str(_dest)

ICON_PATH = str(Path(icons.__file__).parent / "splash.png")


def write_protocol_handler(src, dest):
    # Copy the protocol handler. We use shebangs on Unix systems to easily
    # execute the application
    Path(dest).parent.mkdir(parents=True, exist_ok=True)
    if ON_WINDOWS:
        # Copy the file directly
        shutil.copy(src, dest)
    else:
        # Copy the contents of the file while prepending the shebang
        with open(dest, 'w') as dest_file:
            dest_file.write(PYTHON_SHEBANG + "\n")
            with open(src) as src_file:
                dest_file.writelines(src_file.readlines())
        # Make the file executable
        stat = os.stat(dest)
        os.chmod(dest, stat.st_mode | 0o111)


def register_protocol():
    handler_path = PROTOCOL_HANDLER_DEST
    write_protocol_handler(src=PROTOCOL_HANDLER_SRC,
                           dest=handler_path)
    if system() == "Windows":
        _register_protocol_windows(scheme="karabo", handler=handler_path)
    elif system() == "Linux":
        _register_protocol_linux(scheme="karabo", handler=handler_path)


def _register_protocol_windows(scheme, handler):
    settings = QSettings("HKEY_CURRENT_USER\\Software\\Classes",
                         QSettings.NativeFormat)
    settings.beginGroup(scheme)
    settings.setValue("Default", f"URL:{scheme} Protocol")
    settings.setValue("URL Protocol", "")

    cmd = f'"{sys.executable}" "{handler}" "%1"'
    settings.setValue("shell/open/command/Default", cmd)


def _register_protocol_linux(scheme, handler):
    # Create a desktop file to call the protocol handler
    path = create_linux_desktop_file(
        name="Karabo",
        icon=ICON_PATH,
        command=f"{handler} %u",
        scheme="karabo")
    # Register the scheme on both `gio` and `xdg`
    gio_cmd = ["gio", "mime", f"x-scheme-handler/{scheme}", path.name]
    xdg_cmd = ["xdg-mime", "default", path.name, f"x-scheme-handler/{scheme}"]

    for cmd in (gio_cmd, xdg_cmd):
        try:
            subprocess.call(cmd)
        except (subprocess.CalledProcessError, FileNotFoundError):
            # We are ignoring errors in case one of the two url management
            # is missing.
            RuntimeWarning(f"Scheme registration with {cmd[0]} has failed.")


if __name__ == "__main__":
    register_protocol()
